/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Meisam Vosoughpour
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>

#include <Pass.h>
#include <Degree.h>
#include <Logger.h>
#include <Basics.h>
#include <Defines.h>
#include <Dribble.h>
#include <BasicKick.h>
#include <BasicTurn.h>


//#define DRIBBLE_LOGLEVEL1


using namespace std;

// class Dribble
Dribble::Dribble(const WorldModel *worldModel): AdvancedAction(worldModel)
{
}

Dribble::~Dribble()
{
}

void Dribble::execute(Form &form, const Library &library)
{
	LOG << "Dribble::execute" << endl;

	command = dribbleCommand;
}

Point Dribble::setByPointInformation(float pointBodyAngle, 
		float cautionDistance)
{
	float dribbleAngle = worldModel->getBody().getBodyDir();
	Vector dribbleVector;
	
	Body bodyNoDash;
	bodyNoDash = worldModel->getBody();
	bodyNoDash.simulateByDynamics();

	float kickableArea = worldModel->getBody().getSize() +
			 worldModel->getBody().getKickableMargin() +
			 worldModel->getBall().getSize();

	
	dribbleAngle += pointBodyAngle;
	dribbleAngle = Degree::normalizeAngle(dribbleAngle);	
	Point dribblePoint2;
	dribbleVector.setAsPolar(kickableArea - cautionDistance, dribbleAngle);

	dribblePoint2.x = bodyNoDash.getPos().getX() + dribbleVector.getX();
	dribblePoint2.y = bodyNoDash.getPos().getY() + dribbleVector.getY();
	return dribblePoint2;
}  

void Dribble::setDribblePoints()
{
	float firstXPathStep = 42;


	dribbleTarget.x = 52.5;

	if (worldModel->getBall().getPos().getX() < firstXPathStep)
		dribbleTarget.y = worldModel->getBall().getPos().getY();
	else
		dribbleTarget.y = 0;	

	if (worldModel->getPlayMode() == PM_PENALTY_READY && 
		worldModel->getVirtualPlayMode() == VPM_PENALTY_TAKEN)
	{
		if (worldModel->getBall().getPos().getX() > 33 && 
			worldModel->getBall().getPos().getX() < 42)
		{
			if (worldModel->getBall().getPos().getY() > 0)
			{
				dribbleTarget.x = 52.5;
				dribbleTarget.y = 34;
			}	
			if (worldModel->getBall().getPos().getY() < 0)
			{
				dribbleTarget.x = 52.5;
				dribbleTarget.y = -34;
			}	
		}
	}
	
	idealDribbleTarget = dribbleTarget;			
	isOpponent = 0;
	opponentInSence = 0;
	targetNums = 36;
	cautionLength = 12;
	//man meisam hastam
	angleWeightRate = 1.6;
	craudWeightRate = 8;
	
	for (int i = 0; i < targetNums; i++)
	{
		targetWeights[i] = 0;
		targetAngles[i] = 0;
	}


	angle[0] = 30;
	angle[1] = -30;
	angle[2] = 70;
	angle[3] = -70;
	angle[4] = 15;
	angle[5] = -15;
	angle[6] = 105;
	angle[7] = -105;
	angle[8] = 135;
	angle[9] = -135;
	angle[10] = 165;
	angle[11] = -165;

	virtualWeight[0] = 6;
	virtualWeight[1] = 6;
	virtualWeight[2] = 5;
	virtualWeight[3] = 5;
	virtualWeight[4] = 4;
	virtualWeight[5] = 4;
	virtualWeight[6] = 3;
	virtualWeight[7] = 3;
	virtualWeight[8] = 2;
	virtualWeight[9] = 2;
	virtualWeight[10] = 1;
	virtualWeight[11] = 1;
	
	noise1[0] = .1;
	noise1[1] = .1;
	noise1[2] = .3;
	noise1[3] = .3;
	noise1[4] = .1;
	noise1[5] = .1;
	noise1[6] = .3;
	noise1[7] = .3;
	noise1[8] = .2;
	noise1[9] = .2;
	noise1[10] = .2;
	noise1[11] = .2;


	noise2[0] = .1;
	noise2[1] = .1;
	noise2[2] = .4;
	noise2[3] = .4;
	noise2[4] = .3;
	noise2[5] = .3;
	noise2[6] = .4;
	noise2[7] = .4;
	noise2[8] = .3;
	noise2[9] = .3;
	noise2[10] = .5;
	noise2[11] = .5;
	if (worldModel->getBall().getVel().getMagnitude() > 1.2)
	{
		noise1[0] = .3;
		noise1[1] = .3;
		noise1[2] = .5;
		noise1[3] = .5;
		noise1[4] = .3;
		noise1[5] = .3;
		noise1[6] = .5;
		noise1[7] = .5;
		noise1[8] = .4;
		noise1[9] = .4;
		noise1[10] = .4;
		noise1[11] = .4;
	
	
		noise2[0] = .5;
		noise2[1] = .5;
		noise2[2] = .5;
		noise2[3] = .5;
		noise2[4] = .5;
		noise2[5] = .5;
		noise2[6] = .5;
		noise2[7] = .5;
		noise2[8] = .5;
		noise2[9] = .5;
		noise2[10] = .5;
		noise2[11] = .5;
	}


	dashPower[0] = 100;
	dashPower[1] = 100;
	dashPower[2] = 80;
	dashPower[3] = 80;
	dashPower[4] = 100;
	dashPower[5] = 100;
	dashPower[6] = 60;
	dashPower[7] = 60;
	dashPower[8] = 60;
	dashPower[9] = 60;
	dashPower[10] = 0;
	dashPower[11] = 0;



	maxDist[0] = 1.6;
	maxDist[1] = 1.6;
	maxDist[2] = .3;
	maxDist[3] = .3;
	maxDist[4] = .7;
	maxDist[5] = .7;
	maxDist[6] = .4;
	maxDist[7] = .4;
	maxDist[8] = .4;
	maxDist[9] = .4;
	maxDist[10] = .4;
	maxDist[11] = .4;


	cycleNums[0] = 1;
	cycleNums[1] = 1;
	cycleNums[2] = 1;
	cycleNums[3] = 1;
	cycleNums[4] = 1;
	cycleNums[5] = 1;
	cycleNums[6] = 1;
	cycleNums[7] = 1;
	cycleNums[8] = 1;
	cycleNums[9] = 1;
	cycleNums[10] = 1;
	cycleNums[11] = 1;



	kickableNoise[0] = .1;
	kickableNoise[1] = .1;
	kickableNoise[2] = .2;
	kickableNoise[3] = .2;
	kickableNoise[4] = .1;
	kickableNoise[5] = .1;
	kickableNoise[6] = .2;
	kickableNoise[7] = .2;
	kickableNoise[8] = .1;
	kickableNoise[9] = .1;
	kickableNoise[10] = .1;
	kickableNoise[11] = .1;


	angleNoise[0] = 180;
	angleNoise[1] = 180;
	angleNoise[2] = 70;
	angleNoise[3] = 70;
	angleNoise[4] = 180;
	angleNoise[5] = 70;
	angleNoise[6] = 70;
	angleNoise[7] = 180;
	angleNoise[8] = 180;
	angleNoise[9] = 180;
	angleNoise[10] = 180;
	angleNoise[11] = 180;
	

	kickPower[0] = .58;
	kickPower[1] = .58;
	kickPower[2] = .9;
	kickPower[3] = .9;
	kickPower[4] = .9;
	kickPower[5] = .9;
	kickPower[6] = .9;
	kickPower[7] = .9;
	kickPower[8] = .9;
	kickPower[9] = .9;
	kickPower[10] = .9;
	kickPower[11] = .9;

	
	for (int i = 0; i < 12; i++)
	{
		dribblePoints[i] = setByPointInformation(angle[i], noise1[i]); 
		virtualDribblePoints[i] = setByPointInformation(angle[i], noise1[i]); 
	}
	
	for (int i = 0; i < 12; i++)
		dribbleWeights[i] = 0;
}

void Dribble::setDribbleArgumentsByOpponent()
{

	
	float firstXPathStep = 42;


	dribbleTarget.x = 52.5;

	if (worldModel->getBall().getPos().getX() < firstXPathStep)
		dribbleTarget.y = worldModel->getBall().getPos().getY();
	else
		dribbleTarget.y = 0;	


	if (worldModel->getPlayMode() == PM_PENALTY_READY && 
		worldModel->getVirtualPlayMode() == VPM_PENALTY_TAKEN)
	{
		if (worldModel->getBall().getPos().getX() > 33 && 
			worldModel->getBall().getPos().getX() < 42)
		{
			if (worldModel->getBall().getPos().getY() > 0)
			{
				dribbleTarget.x = 52.5;
				dribbleTarget.y = 34;
			}	
			if (worldModel->getBall().getPos().getY() < 0)
			{
				dribbleTarget.x = 52.5;
				dribbleTarget.y = -34;
			}	
		}
	}
	
	idealDribbleTarget = dribbleTarget;	
	isOpponent = 0;
	targetNums = 36;
	cautionLength = 12;
	angleWeightRate = 1.6;
	craudWeightRate = 8;
	
	for (int i = 0; i < targetNums; i++)
	{
		targetWeights[i] = 0;
		targetAngles[i] = 0;
	}
	
	angle[0] = 30;
	angle[1] = -30;
	angle[2] = 70;
	angle[3] = -70;
	angle[4] = 15;
	angle[5] = -15;
	angle[6] = 105;
	angle[7] = -105;
	angle[8] = 135;
	angle[9] = -135;
	angle[10] = 165;
	angle[11] = -165;

	virtualWeight[0] = 6;
	virtualWeight[1] = 6;
	virtualWeight[2] = 5;
	virtualWeight[3] = 5;
	virtualWeight[4] = 4;
	virtualWeight[5] = 4;
	virtualWeight[6] = 3;
	virtualWeight[7] = 3;
	virtualWeight[8] = 2;
	virtualWeight[9] = 2;
	virtualWeight[10] = 1;
	virtualWeight[11] = 1;
	
	noise1[0] = .1;
	noise1[1] = .1;
	noise1[2] = .2;
	noise1[3] = .2;
	noise1[4] = .1;
	noise1[5] = .1;
	noise1[6] = .2;
	noise1[7] = .2;
	noise1[8] = .1;
	noise1[9] = .1;
	noise1[10] = .1;
	noise1[11] = .1;


	noise2[0] = .3;
	noise2[1] = .3;
	noise2[2] = .4;
	noise2[3] = .4;
	noise2[4] = .3;
	noise2[5] = .3;
	noise2[6] = .4;
	noise2[7] = .4;
	noise2[8] = .3;
	noise2[9] = .3;
	noise2[10] = .4;
	noise2[11] = .4;


	dashPower[0] = 100;
	dashPower[1] = 100;
	dashPower[2] = 80;
	dashPower[3] = 80;
	dashPower[4] = 100;
	dashPower[5] = 100;
	dashPower[6] = 60;
	dashPower[7] = 60;
	dashPower[8] = 60;
	dashPower[9] = 60;
	dashPower[10] = 50;
	dashPower[11] = 50;

	maxDist[0] = 0.4;
	maxDist[1] = 0.4;
	maxDist[2] = 0;
	maxDist[3] = 0;
	maxDist[4] = 0;
	maxDist[5] = 0;
	maxDist[6] = 0;
	maxDist[7] = 0;
	maxDist[8] = 0;
	maxDist[9] = 0;
	maxDist[10] = 0;
	maxDist[11] = 0;

	cycleNums[0] = 1;
	cycleNums[1] = 1;
	cycleNums[2] = 1;
	cycleNums[3] = 1;
	cycleNums[4] = 1;
	cycleNums[5] = 1;
	cycleNums[6] = 1;
	cycleNums[7] = 1;
	cycleNums[8] = 1;
	cycleNums[9] = 1;
	cycleNums[10] = 1;
	cycleNums[11] = 1;

	kickableNoise[0] = .1;
	kickableNoise[1] = .1;
	kickableNoise[2] = .2;
	kickableNoise[3] = .2;
	kickableNoise[4] = .1;
	kickableNoise[5] = .1;
	kickableNoise[6] = .2;
	kickableNoise[7] = .2;
	kickableNoise[8] = .1;
	kickableNoise[9] = .1;
	kickableNoise[10] = .1;
	kickableNoise[11] = .1;

	angleNoise[0] = 180;
	angleNoise[1] = 180;
	angleNoise[2] = 70;
	angleNoise[3] = 70;
	angleNoise[4] = 180;
	angleNoise[5] = 70;
	angleNoise[6] = 70;
	angleNoise[7] = 180;
	angleNoise[8] = 180;
	angleNoise[9] = 180;
	angleNoise[10] = 180;
	angleNoise[11] = 180;

	kickPower[0] = 1;
	kickPower[1] = 1;
	kickPower[2] = 1;
	kickPower[3] = 1;
	kickPower[4] = 1;
	kickPower[5] = 1;
	kickPower[6] = 1;
	kickPower[7] = 1;
	kickPower[8] = 1;
	kickPower[9] = 1;
	kickPower[10] = 1;
	kickPower[11] = 1;
	
	
	for (int i = 0; i < 12; i++)
	{
		dribblePoints[i] = setByPointInformation(angle[i], noise1[i]); 
		virtualDribblePoints[i] = setByPointInformation(angle[i], noise1[i]); 
	}
	
	for (int i = 0; i < 12; i++)
		dribbleWeights[i] = 0;
}

void Dribble::setOpponentInWeight(float senceArea)
{
	opponentInSence = 0;
	Point playerPoint;
	for (int i = 0; i < 12; i++)
	{
		for (int j = 0; j < FULL_PLAYERS_NUM; j++)
		{
	/*BY AHMAD*/    if (worldModel->getFullPlayer(TID_TEAMMATE, j).isValid() &&
				!worldModel->getFullPlayer(TID_TEAMMATE, j).isBody())
			{
				playerPoint = worldModel->getFullPlayer(TID_TEAMMATE, j).getPos();
				
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}
	
			if (worldModel->getFullPlayer(TID_OPPONENT, j).isValid())
			{
				playerPoint = worldModel->getFullPlayer(TID_OPPONENT, j).getPos();
				
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}	
		}
		for (int j = 0; j < HALF_PLAYERS_NUM; j++)
		{
	/*BY AHMAD*/   if (worldModel->getHalfPlayer(TID_TEAMMATE, j).isValid())
			{
				playerPoint = worldModel->getHalfPlayer(TID_TEAMMATE, j).getPos();
				
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}

			if (worldModel->getHalfPlayer(TID_OPPONENT, j).isValid())
			{
				playerPoint = worldModel->getHalfPlayer(TID_OPPONENT, j).getPos();

				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);

				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}	
		}

		for (int j = 0; j < QUARTER_PLAYERS_NUM; j++)
		{
			if (worldModel->getQuarterPlayer(j).isAlive())
			{
				playerPoint = worldModel->getQuarterPlayer(j).getPos();
	
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}	
		}
	}
}

int Dribble::bestDribblePoint()
{
	float senceArea = 2; // Sassan (1.8)
	opponentInSence = 0;
	setOpponentInWeight(senceArea);
	if (opponentInSence)
	{
		setDribbleArgumentsByOpponent();
		setOpponentInWeight(senceArea);
	}
	
	int bestDribbleCount = 0;
	float goodWeight = .09;
	
	
	for (int i = 0; i < 12; i++)
	{
#ifdef DRIBBLE_LOGLEVEL1
		LOG << "dribbleWeight[" << i << "]" << dribbleWeights[i] << endl;	
#endif

		if (dribbleWeights[i] > dribbleWeights[bestDribbleCount] || 
			(virtualWeight[i] > virtualWeight[bestDribbleCount] && 
			dribbleWeights[i] + goodWeight > dribbleWeights[bestDribbleCount]))	
			bestDribbleCount = i;
	}
		
	if (bestDribbleCount == 0 && dribbleWeights[0] == dribbleWeights[1])
	{
		Command *dashCommand = new DashCommand(AT_DRIBBLE, 100);
		Ball ballByDash, ballNoDash;
		ballByDash = worldModel->getBall();
		ballNoDash = worldModel->getBall();

		Body bodyByDash, bodyNoDash;
		bodyByDash  = worldModel->getBody();
		bodyNoDash = worldModel->getBody();
	
		bodyNoDash.simulateByDynamics();
		bodyByDash.simulateByAction(dashCommand);
		bodyByDash.simulateByDynamics();
		ballNoDash.simulateByDynamics(bodyNoDash);
		ballByDash.simulateByDynamics(bodyByDash);
		if (ballNoDash.getBodyVec().getDirection() <= 0)
			bestDribbleCount = 1;
		delete dashCommand;
	}
	return bestDribbleCount;			
}


void Dribble::setAngleWeight(int targetCount)
{
	float angleWeight = 0;
	float deltaAngle = targetAngles[targetCount] - targetAngles[0];
	deltaAngle = Degree::normalizeAngle(deltaAngle);

	Vector tempVector;
	tempVector.setAsPolar(cautionLength, deltaAngle);

	angleWeight += ((180 - fabs(deltaAngle)) / 30.00) * .15;

	float deltaBodyAngle = targetAngles[targetCount] - 
				worldModel->getBody().getBodyDir();
	deltaBodyAngle = Degree::normalizeAngle(deltaBodyAngle);
	
	 
	//KOMAK in if bayad az lahaze algorithmi avaz shavad.
	if (isOpponent && fabs(deltaBodyAngle) < 20 && targetAngles[targetCount]) 
	{
/*	
	float myModel = fabs(targetWeights[targetCount]);
	if (myModel > 80)
		myModel = 80;
	if (myModel < 3)
		myModel  = 3;
*/		
		angleWeight += .04;
	}
	targetWeights[targetCount] += angleWeight * angleWeightRate;
}

void Dribble::setCraudWeight(int targetCount)
{
	float beginAngle = targetAngles[targetCount] - 90;
	beginAngle = Degree::normalizeAngle(beginAngle);
	Vector tempVector;
	
	float currentAngle;
	float craudWeight = 0;
	float playerDist;
	float negWeight = 0;
	float xRate = .25;
	float localNegWeight;
	
	for (int j = 0; j < FULL_PLAYERS_NUM; j++)
		if (worldModel->getFullPlayer(TID_OPPONENT, j).isValid())
		{
			currentAngle = worldModel->getFullPlayer(TID_OPPONENT, j).getAbsVec().getDirection() - beginAngle;	
			currentAngle = Degree::normalizeAngle(currentAngle);

			playerDist = worldModel->getFullPlayer(TID_OPPONENT, j).getBodyVec().getMagnitude();			
			tempVector.setAsPolar(playerDist, currentAngle);

			if (playerDist < cautionLength * 1.2)
				isOpponent = 1;

			if (tempVector.getY() <= cautionLength && tempVector.getY() > 0 && 
					fabs(tempVector.getX()) <= (cautionLength * xRate))
			{
				localNegWeight = -(tempVector.getY() - fabs(tempVector.getX()));
				localNegWeight *= (fabs(tempVector.getX()) / cautionLength) / xRate;

				if (playerDist < 5)
				{
					float rate = 5 - playerDist;
					if (rate < 1.5)
						rate = 1.5;
	
					localNegWeight *= rate;
				}

				if (localNegWeight > 0)
					localNegWeight = 0;

				if (localNegWeight < negWeight)
					negWeight = localNegWeight;
			}
		}	


	for (int j = 0; j < QUARTER_PLAYERS_NUM; j++)
		if (worldModel->getQuarterPlayer(j).isAlive())
		{
			currentAngle = worldModel->getQuarterPlayer(j).getAbsVec().getDirection() - beginAngle;	
			currentAngle = Degree::normalizeAngle(currentAngle);
	
			playerDist = worldModel->getQuarterPlayer(j).getBodyVec().getMagnitude();			
			tempVector.setAsPolar(playerDist, currentAngle);
	
			if (playerDist < cautionLength * 1.2)
				isOpponent = 1;
	
			if (tempVector.getY() <= cautionLength && tempVector.getY() > 0 && 
					fabs(tempVector.getX()) <= (cautionLength * xRate))
			{
				localNegWeight = -(tempVector.getY() - fabs(tempVector.getX()));
				localNegWeight *= (fabs(tempVector.getX()) / cautionLength) / xRate;

				if (playerDist < 5)
				{
					float rate = 5 - playerDist;
					if (rate < 1.5)
						rate = 1.5;

					localNegWeight *= rate;
				}

				if (localNegWeight > 0)
					localNegWeight = 0;

				if (localNegWeight < negWeight)
					negWeight = localNegWeight;
			}
		}
		
	craudWeight = negWeight;
	targetWeights[targetCount] += craudWeight * craudWeightRate;	
}

void Dribble::setTargetWeight(int targetCount)
{
	setCraudWeight(targetCount);
	setAngleWeight(targetCount);
//	LOG << "TotalWeight : " << targetAngles[targetCount] << " " << 
//		targetWeights[targetCount] << endl; 
}


void Dribble::setDribbleTarget(int dribbleCount)
{
	Vector tempVector;
	
	tempVector.setByPoints(Point(worldModel->getBody().getPos().getX(), 
			worldModel->getBody().getPos().getY()), dribbleTarget);
	
	float beginAngle = tempVector.getDirection();   
	
	for (int i = 0; i < targetNums; i++) 
	{
		targetAngles[i] = beginAngle + (i * 10); 	
		targetAngles[i] = Degree::normalizeAngle(targetAngles[i]);
		setTargetWeight(i);
	}

	if (worldModel->getBall().getPos().getY() >= 31)
		for (int i = 0; i < targetNums; i++)
			if (targetAngles[i] > 0 || (worldModel->getBall().getPos().getY() >= 
				32 && targetAngles[i] >= 0))
				targetWeights[i] -= 1000;

	if (worldModel->getBall().getPos().getY() <= -31)
		for (int i = 0; i < targetNums; i++)
			if (targetAngles[i] < 0 || (worldModel->getBall().getPos().getY() <= 
				-32 && targetAngles[i] <= 0))
				targetWeights[i] -= 1000;
	
	if (worldModel->getBall().getPos().getX() >= 50)
		for (int i = 0; i < targetNums; i++)
			if (fabs(targetAngles[i]) < 90 || 
				(worldModel->getBall().getPos().getX() >= 51 && 
				fabs(targetAngles[i]) <= 90))
				targetWeights[i] -= 1000;

	int bestDribbleTarget = 0;

	for (int i = 0; i < targetNums; i++)
		if (targetWeights[i] > targetWeights[bestDribbleTarget])
			bestDribbleTarget = i;

	Vector lastVector;
	lastVector.setAsPolar(10, targetAngles[bestDribbleTarget]);

	dribbleTarget.x = worldModel->getBody().getPos().getX() + lastVector.getX();
	dribbleTarget.y = worldModel->getBody().getPos().getY() + lastVector.getY();

#ifdef DRIBBLE_LOGLEVEL1
	LOG << "MainDribbletarget : " << dribbleTarget << endl;
#endif
}

bool Dribble::checkDribbleTarget(int dribbleCount)
{
	float kickableArea;
	float deltaAngle;
	float dribbleAngle;
	float maxDeltaAngle;
	Vector targetVector;
	Vector dribbleVector;
	Point ballPoint;
	Point dribblePoint;

	setDribbleTarget(dribbleCount);

	kickableArea = worldModel->getBody().getSize() +
			 worldModel->getBody().getKickableMargin() +
			 worldModel->getBall().getSize();
	
	targetVector.setByPoints(Point(worldModel->getBody().getPos().getX(), 
			worldModel->getBody().getPos().getY()), dribbleTarget);

	deltaAngle = targetVector.getDirection() - 
			worldModel->getBody().getBodyDir();
	deltaAngle = Degree::normalizeAngle(deltaAngle);
	
	ballPoint.x = worldModel->getBall().getPos().getX();
	ballPoint.y = worldModel->getBall().getPos().getY();
	
	maxDeltaAngle = 30;
	if (worldModel->getBody().getPos().getY() > 31 && 
		worldModel->getBody().getBodyDir() > 0 && deltaAngle < 0)  
		maxDeltaAngle = 0;				   
	
	if (worldModel->getBody().getPos().getY() < -31 && 
		worldModel->getBody().getBodyDir() < 0 && deltaAngle > 0) 
		maxDeltaAngle = 0;				   
	
	if (worldModel->getBody().getPos().getX() > 50 && 
		fabs(worldModel->getBody().getBodyDir()) < 90) 
		maxDeltaAngle = 0;				   

	dribbleAngle = worldModel->getBody().getBodyDir();
	
	dribbleAngle += angle[dribbleCount];
	dribbleAngle = Degree::normalizeAngle(dribbleAngle);	

	dribbleVector.setAsPolar(kickableArea - noise1[dribbleCount] - .2, dribbleAngle);
	
	dribblePoint.x = checkTargetBody.getPos().getX() + dribbleVector.getX();
	dribblePoint.y = checkTargetBody.getPos().getY() + dribbleVector.getY();
	
	float extraNoise = 0;
	
	if (worldModel->getBall().getVel().getMagnitude() > .5)
		extraNoise = .1;
		
	if (fabs(worldModel->getBall().getPos().getX()) > 49.5 ||
		fabs(worldModel->getBall().getPos().getY()) > 32)
		extraNoise += .25;	

	if (fabs(deltaAngle) > maxDeltaAngle)
	{
		if (checkTargetBall.getBodyVec().getMagnitude() + 
			kickableNoise[dribbleCount] + extraNoise >= kickableArea || 
			(opponentInSence && hypot(dribblePoint.x - ballPoint.x, 
			dribblePoint.y - ballPoint.y) > 
			max((maxDist[dribbleCount] * kickableArea), (float)(.2))))
		{
			dribbleCommand = KickToPointCycles(AT_DRIBBLE, dribblePoint, 1, 
				worldModel->getBall(), worldModel->getBody()).getCommand();  
			
#ifdef DRIBBLE_LOGLEVEL1
			LOG << "dribbleBUG 1 : dribblePoint : " << dribblePoint << endl;
#endif
		}
		else
			dribbleCommand = TurnAngle(AT_DRIBBLE, deltaAngle, 
				worldModel->getBody()).getCommand();
		return 0;
	}
	return 1;
}


float  Dribble::getInlineDribbleWeight(int dribbleCount, float senceArea)
{
	float inlineDribbleWeight = 0;
	Point playerPoint;
	for (int j = 0; j < FULL_PLAYERS_NUM; j++)
	{
/*	BY AHMAD    if (worldModel->getFullPlayer(TID_TEAMMATE, j).isValid() &&
				!worldModel->getFullPlayer(TID_TEAMMATE, j).isBody())
			{
				playerPoint = worldModel->getFullPlayer(TID_TEAMMATE, j).getPos();
				
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}	*/
	
		if (worldModel->getFullPlayer(TID_OPPONENT, j).isValid())
		{
			playerPoint = worldModel->getFullPlayer(TID_OPPONENT, j).getPos();
			
			float playerDist =  hypot(dribblePoints[dribbleCount].x - playerPoint.x, 
					dribblePoints[dribbleCount].y - playerPoint.y);
			if (playerDist < senceArea)
			{
				inlineDribbleWeight -= senceArea - playerDist;	
				opponentInSence = 1;
 			}
		}	
	}
	for (int j = 0; j < HALF_PLAYERS_NUM; j++)
	{
/*	BY AHMAD   if (worldModel->getHalfPlayer(TID_TEAMMATE, j).isValid())
			{
				playerPoint = worldModel->getHalfPlayer(TID_TEAMMATE, j).getPos();
				
				float playerDist =  hypot(dribblePoints[i].x - playerPoint.x, 
						dribblePoints[i].y - playerPoint.y);
	
				if (playerDist < senceArea)
				{
					dribbleWeights[i] -= senceArea - playerDist;	
					opponentInSence = 1;
 				}
			}	*/

		if (worldModel->getHalfPlayer(TID_OPPONENT, j).isValid())
		{
			playerPoint = worldModel->getHalfPlayer(TID_OPPONENT, j).getPos();

			float playerDist =  hypot(dribblePoints[dribbleCount].x - 
				playerPoint.x, dribblePoints[dribbleCount].y - playerPoint.y);

			if (playerDist < senceArea)
				inlineDribbleWeight -= senceArea - playerDist;	
		}	
	}

	for (int j = 0; j < QUARTER_PLAYERS_NUM; j++)
	{
		if (worldModel->getQuarterPlayer(j).isAlive())
		{
			playerPoint = worldModel->getQuarterPlayer(j).getPos();
	
			float playerDist =  hypot(dribblePoints[dribbleCount].x - playerPoint.x, 
					dribblePoints[dribbleCount].y - playerPoint.y);
	
			if (playerDist < senceArea)
				inlineDribbleWeight -= senceArea - playerDist;	
		}	
	}
	return inlineDribbleWeight;
}

bool Dribble::ballInOppKickable()
{
	Point playerPoint;
	for (int j = 0; j < FULL_PLAYERS_NUM; j++)
		if (worldModel->getFullPlayer(TID_OPPONENT, j).isValid())
		{
			playerPoint = worldModel->getFullPlayer(TID_OPPONENT, j).getPos();

			float playerDist =  hypot(checkTargetBall.getPos().getX() - playerPoint.x, 
				checkTargetBall.getPos().getY() - playerPoint.y);

			if (playerDist <= 1.2)
				return 1;
		}	
	
	for (int j = 0; j < QUARTER_PLAYERS_NUM; j++)
		if (worldModel->getQuarterPlayer(j).isAlive())
		{
			playerPoint = worldModel->getQuarterPlayer(j).getPos();	

			float playerDist =  hypot(checkTargetBall.getPos().getX() - playerPoint.x, 
					checkTargetBall.getPos().getY() - playerPoint.y);	

			if (playerDist <= 1.2)
				return 1;
		}

	return 0;
}


float Dribble::getValue(const Library &library)
{
	LOG << "Dribble::getValue" << endl;

	int dribbleCount;
	int localCycleNums;
	float kickableArea;
	float localDashPower;
	Point dribblePoint;
	Point ballPoint;
	Vector dribbleVector;
	Ball ballByDash, ballNoDash, nextBall;
	Body bodyByDash, bodyNoDash, nextBody;
		
	setDribblePoints();
	dribbleCount = bestDribblePoint();
	kickableArea = worldModel->getBody().getSize() +
			 worldModel->getBody().getKickableMargin() +
			 worldModel->getBall().getSize();
	
	ballPoint.x = worldModel->getBall().getPos().getX();
	ballPoint.y = worldModel->getBall().getPos().getY();

	float dribbleAngle = worldModel->getBody().getBodyDir();
	
	dribbleAngle += angle[dribbleCount];
	dribbleAngle = Degree::normalizeAngle(dribbleAngle);	

	dribbleVector.setAsPolar(kickableArea - noise1[dribbleCount], dribbleAngle);
	
	ballByDash = worldModel->getBall();
	ballNoDash = worldModel->getBall();
	nextBall = worldModel->getBall();

	bodyByDash  = worldModel->getBody();
	bodyNoDash = worldModel->getBody();
	nextBody = worldModel->getBody();

	localDashPower = dashPower[dribbleCount];
	
	if (worldModel->getBody().getStamina() < 1400)
		localDashPower = worldModel->getBody().getStaminaIncMax();

	Command *dashCommand = new DashCommand(AT_DRIBBLE, localDashPower);
	
	bodyNoDash.simulateByDynamics();
	bodyByDash.simulateByAction(dashCommand);
	bodyByDash.simulateByDynamics();
	
	ballNoDash.simulateByDynamics(bodyNoDash);
	ballByDash.simulateByDynamics(bodyByDash);
  
  	float ballDistByDash, ballDistNoDash;
	  
	ballDistByDash = hypot(ballByDash.getPos().getX() - 
		ballByDash.getPos().getX(), ballByDash.getPos().getY() - 
		ballByDash.getPos().getY()); 
		
	ballDistNoDash = hypot(ballNoDash.getPos().getX() - 
		ballNoDash.getPos().getX(), ballNoDash.getPos().getY() - 
		ballNoDash.getPos().getY()); 
	
	dribblePoint.x = bodyNoDash.getPos().getX() + dribbleVector.getX();
	dribblePoint.y = bodyNoDash.getPos().getY() + dribbleVector.getY();

	Point dribblePoint2;

	if (dribbleCount == 0 || dribbleCount == 1)
	{
		dribbleAngle = worldModel->getBody().getBodyDir();
	
		dribbleAngle += angle[dribbleCount];
		dribbleAngle = Degree::normalizeAngle(dribbleAngle);	

		dribbleVector.setAsPolar(kickableArea - noise2[dribbleCount] , 
			dribbleAngle);

		dribblePoint.x = bodyNoDash.getPos().getX() + dribbleVector.getX();
		dribblePoint.y = bodyNoDash.getPos().getY() + dribbleVector.getY();
		
	
		dribblePoint2.x = worldModel->getBall().getPos().getX() + 
			(dribblePoint.x - worldModel->getBall().getPos().getX()) * 
			kickPower[dribbleCount];
		dribblePoint2.y = worldModel->getBall().getPos().getY() + 
			(dribblePoint.y - worldModel->getBall().getPos().getY()) * 
			kickPower[dribbleCount];
	
#ifdef DRIBBLE_LOGLEVEL1
		LOG << "inja dribble ast : dribblePoint : " << dribblePoint << endl;
		LOG << "inja dribble ast : dribblePoint2 : " << dribblePoint2 << endl;
		LOG << "dribbleVector : " << dribbleVector.getMagnitude() << " " << dribbleVector.getDirection() << endl;
#endif
				
//		LOG << "dribblePoint2 : Before : " << dribblePoint2.x << " " << dribblePoint2.y << endl;
	
		float dribblePointDist = hypot(bodyNoDash.getPos().getX() - 
			dribblePoint2.x, bodyNoDash.getPos().getY() - dribblePoint2.y) + .1; 
		
		if ((dribblePointDist) >= kickableArea) 
			kickPower[dribbleCount] = 1;
	
		dribblePoint2.x = worldModel->getBall().getPos().getX() + 
			(dribblePoint.x - worldModel->getBall().getPos().getX()) * 
			kickPower[dribbleCount];
		dribblePoint2.y = worldModel->getBall().getPos().getY() + 
			(dribblePoint.y - worldModel->getBall().getPos().getY()) * 
			kickPower[dribbleCount];

//		LOG << "dribblePoint2 : After : " << dribblePoint2 << endl;
	}
	else
	{
		dribbleAngle = worldModel->getBody().getBodyDir();
	
		dribbleAngle += angle[dribbleCount];
		dribbleAngle = Degree::normalizeAngle(dribbleAngle);	
	
		dribbleVector.setAsPolar(kickableArea - noise2[dribbleCount], dribbleAngle);

		dribblePoint2.x = bodyNoDash.getPos().getX() + dribbleVector.getX();
		dribblePoint2.y = bodyNoDash.getPos().getY() + dribbleVector.getY();
	}
	
	localCycleNums = cycleNums[dribbleCount];

	Vector ballVector;
	ballVector.setByPoints(Point(ballNoDash.getPos().getX(), 
		ballNoDash.getPos().getY()), dribblePoint);

	float deltaAngle;
	deltaAngle = ballVector.getDirection() - worldModel->getBody().getBodyDir();
	deltaAngle = Degree::normalizeAngle(deltaAngle);

	if (fabs(deltaAngle) > angleNoise[dribbleCount] && 
		ballNoDash.getBodyVec().getMagnitude() + 
		kickableNoise[dribbleCount] < kickableArea)
		localCycleNums = 2;	 

	///LOG << "dribblePoint: " << dribblePoint << endl;
	///LOG << "angle: " << angle[dribbleCount] << endl;
	///LOG << "noise1: " << noise1[dribbleCount] << endl;
	///LOG << "noise2: " << noise2[dribbleCount] << endl;
	///LOG << "angleNoise: " << angleNoise[dribbleCount] << endl;
	///LOG << "maxDist: " << maxDist[dribbleCount] << endl;
	///LOG << "kickableNoise: " << kickableNoise[dribbleCount] << endl;
	///LOG << "Dribble : cycleNums : " << localCycleNums << endl;	

	if (ballInOppKickable())
		maxDist[dribbleCount] = 0;
		
	float collisionDist;
	collisionDist = hypot(bodyByDash.getPos().getX() - 
		ballByDash.getPos().getX(), bodyByDash.getPos().getY() - 
		ballByDash.getPos().getY());

#ifdef DRIBBLE_LOGLEVEL1
//	LOG << "collisionDist : " << collisionDist << endl;
	LOG << "dribbleCount : " << dribbleCount << endl;
//	LOG << "maxDist[dribbleCount] : " << maxDist[dribbleCount] << endl;
//	LOG << "ballByDash.getBodyVec().getMagnitude() : " << 
//		ballByDash.getBodyVec().getMagnitude() << endl;
//	LOG << "kickableNoise[dribbleCount] : " << kickableNoise[dribbleCount] << endl;
//	LOG << "kickableArea : " << kickableArea << endl;
	LOG << "OpponentInSense : " << opponentInSence << endl;
//	LOG << "kickPower[dribbleCount] : " << kickPower[dribbleCount] << endl;
#endif


#ifdef DRIBBLE_LOGLEVEL1
/////////////////////////NEW LOGS///////////////////////
	LOG << "angle[DC]         : " << angle[dribbleCount] << endl;
	LOG << "virtualWeight[DC] : " << virtualWeight[dribbleCount] << endl;
	LOG << "noise1[DC]        : " << noise1[dribbleCount] << endl;
	LOG << "noise2[DC]        : " << noise2[dribbleCount] << endl;
	LOG << "dashPower[DC]     : " << dashPower[dribbleCount] << endl;
	LOG << "maxDist[DC]       : " << maxDist[dribbleCount] << endl;
	LOG << "cycleNums[DC]     : " << cycleNums[dribbleCount] << endl;
	LOG << "kickableNoise[DC] : " << kickableNoise[dribbleCount] << endl;
	LOG << "angleNoise[DC]    : " << angleNoise[dribbleCount] << endl;
	LOG << "kickPower[DC]     : " << kickPower[dribbleCount] << endl;
	LOG << "kickableArea      : " << kickableArea << endl;
/////////////////////////NEW LOGS END///////////////////
#endif


	
	if (ballNoDash.getBodyVec().getMagnitude() + .1 >= kickableArea)
	{
		if (ballByDash.getBodyVec().getMagnitude() + 
			kickableNoise[dribbleCount] < kickableArea && 
			hypot(dribblePoint.x - ballPoint.x, dribblePoint.y - ballPoint.y) < 
			(maxDist[dribbleCount] * kickableArea))
		{
//			delete dribbleCommand;
#ifdef DRIBBLE_LOGLEVEL1
			LOG << "Dribble : Status1" << endl;
#endif
			dribbleCommand = new DashCommand(AT_DRIBBLE, localDashPower);
			
			nextBody.simulateByAction(dribbleCommand);
			nextBody.simulateByDynamics();
			nextBall.simulateByDynamics(nextBody);
		}	
		else
		{
			dribbleCommand = KickToPointCycles(AT_DRIBBLE, dribblePoint2, 
				localCycleNums, worldModel->getBall(), 
				worldModel->getBody()).getCommand();			
				
			nextBody.simulateByDynamics();
			nextBall.simulateByAction(nextBody, dribbleCommand);
			nextBall.simulateByDynamics(nextBody);

#ifdef DRIBBLE_LOGLEVEL1
			LOG << "Dribble : Status2" << endl;		   
			LOG << "dribblePoint2 : " << dribblePoint2 << endl;
//			LOG << "dribblePoint2 of Selected dribbleCount :::::::: : " << 
//				drbblePoint2 << endl;
//			LOG << "DribblePoint1 of Selected dribbleCount :::::::: : " << 
//				dribblePoint << endl;
#endif
		}
	}
	else
	{
		if (hypot(dribblePoint.x - ballPoint.x, dribblePoint.y - ballPoint.y) > 
			(maxDist[dribbleCount] * kickableArea)
			|| ballByDash.getBodyVec().getMagnitude() + .1 >= kickableArea
			|| collisionDist <= .4)
		{
			dribbleCommand = KickToPointCycles(AT_DRIBBLE, dribblePoint2, 
				localCycleNums, worldModel->getBall(), 
				worldModel->getBody()).getCommand();

#ifdef DRIBBLE_LOGLEVEL1
			LOG << "Dribble : Status3" << endl;
			LOG << "dribblePoint2 : " << dribblePoint2 << endl;
#endif

			nextBody.simulateByDynamics();
			nextBall.simulateByAction(nextBody, dribbleCommand);
			nextBall.simulateByDynamics(nextBody);
		}
		else 
		{
#ifdef DRIBBLE_LOGLEVEL1
			LOG << "Dribble : Status4" << endl;
#endif

//			delete dribbleCommand;
			dribbleCommand = new DashCommand(AT_DRIBBLE, localDashPower);
			nextBody.simulateByAction(dribbleCommand);
			nextBody.simulateByDynamics();
			nextBall.simulateByDynamics(nextBody);
		}
	}

	checkTargetBall = ballNoDash;
	checkTargetBody = bodyNoDash;
	checkDribbleTarget(dribbleCount);

//	LOG << "dribbleCount Latest : " << dribbleCount << endl;
	
	float dribbleWeight;
	dribbleWeight = 35 - ((dribbleCount / 2) * 11);

	float pathWeight = 0;
	float deltaPathAngle;
	Vector currentPathVector, idealPathVector;

	currentPathVector.setByPoints(Point(worldModel->getBody().getPos().getX(), 
		worldModel->getBody().getPos().getY()), dribbleTarget);

	idealPathVector.setByPoints(Point(worldModel->getBody().getPos().getX(), 
		worldModel->getBody().getPos().getY()), idealDribbleTarget);

	deltaPathAngle = currentPathVector.getDirection() - idealPathVector.getDirection();
	deltaPathAngle = Degree::normalizeAngle(deltaPathAngle);
	deltaPathAngle = fabs(deltaPathAngle);
	
	if (deltaPathAngle < 40)
		pathWeight = (40.00 - deltaPathAngle) / 1.00;
	
	if (dribbleCount == 4 || dribbleCount == 5)
		dribbleWeight -= 20;
	
	dribbleWeight += pathWeight;	
	
	float inlineDribbleWeight = 0;
	inlineDribbleWeight = getInlineDribbleWeight(dribbleCount, 2.8);
	inlineDribbleWeight = pow((inlineDribbleWeight * 3.6), 2) * 3;
	LOG << "inlineDribbleWeight : " << inlineDribbleWeight << endl;
	dribbleWeight -= inlineDribbleWeight;

	delete dashCommand;	
	return dribbleWeight + getBlockedWeight() + getBodyCrossWeight() + 5; //MEISAM2005
}

float Dribble::getBlockedWeight()
{
	float min = DRIBBLE_BLOCK_MAX_WEIGHT;

	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		if (!worldModel->getFullPlayer(TID_OPPONENT, i).isValid())
			continue;

		float cur = getPlayerBlockWeight(
				worldModel->getFullPlayer(TID_OPPONENT, i));
		if (cur < min)
			min = cur;
	}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
	{
		if (!worldModel->getHalfPlayer(TID_OPPONENT, i).isValid())
			continue;

		float cur = getPlayerBlockWeight(
				worldModel->getHalfPlayer(TID_OPPONENT, i));
		if (cur < min)
			min = cur;
	}

	for (unsigned i = 0; i < QUARTER_PLAYERS_NUM; i++)
	{
		if (!worldModel->getQuarterPlayer(i).isValid())
			continue;

		float cur = getPlayerBlockWeight(
				worldModel->getQuarterPlayer(i));
		if (cur < min)
			min = cur;
	}

	LOG << "DribbleBlockedWeight: " << min << endl;
	return min;
}

float Dribble::getPlayerBlockWeight(const Player &player)
{
	if (player.getAbsVec().getMagnitude() > 5)
		return DRIBBLE_BLOCK_MAX_WEIGHT - 35;

LOG << "Dribble::getPlayerBlockWeight" << endl;
LOG << "Player: " << player.getTeamId() << " " << player.getUniNum() << endl;

	if (player.getAbsVec().getMagnitude() < 2.5 &&
		fabs(player.getBodyVec().getDirection()) > 125)
	{
		LOG << "MIN_WEIGHT: in forbidden area." << endl;
		return 0 - 35;
	}

	float deltaAngle;

	if (player.getVel().getMagnitude() > 0.2)
	{
		float bodyDir =
				Degree::normalizeAngle(180 + player.getAbsVec().getDirection());
		deltaAngle = fabs(Degree::getDeltaAngle(bodyDir,
				player.getVel().getDirection()));
	}
	else
		deltaAngle = 180;

LOG << "deltaAngle: " << deltaAngle << endl;

	float distValue = /*sqrt(*/player.getAbsVec().getMagnitude()/*)*/;

LOG << "distValue: " << distValue << endl;

	float distWeight = Basics::reRate(distValue, /*1.20, SQRT_5*/ 2, 5, 0,
			DRIBBLE_BLOCK_MAX_WEIGHT * DRIBBLE_BLOCK_DIST_RATE, RRM_DIRECT);

LOG << "dist from 1.2 to " << SQRT_5 << endl;
LOG << "distWeight: " << distWeight << endl;

	float dirWeight = Basics::reRate(deltaAngle, 20, 80, 0,
			DRIBBLE_BLOCK_MAX_WEIGHT * DRIBBLE_BLOCK_DIR_RATE, RRM_DIRECT);

LOG << "dirWeight: " << dirWeight << endl;
LOG << "Weight: " << distWeight + dirWeight - 35 << endl;

	return distWeight + dirWeight - 35;
}

float Dribble::getBodyCrossWeight()
{
	float shootDir = Pass::getShootDirValue(
			worldModel->getBody().getPos(), worldModel);
	float shootDist = hypot(worldModel->getBody().getPos().getX() - 52.5,
			fmin(worldModel->getBody().getPos().getY() - 4,
				 worldModel->getBody().getPos().getY() + 4));

	float shootValue = shootDir - fmax(shootDist - 10, 0) * 1.15;

	float shootWeight = fmax (0, shootValue * 2);
	LOG << "DribbleShootWeight: " << shootWeight << endl;

	return shootWeight;
}


// class DribblePath
DribblePath::DribblePath(const WorldModel *worldModel):	worldModel(worldModel)
{
}

void DribblePath::setAngleWeight(int targetCount)
{
	float angleWeight = 0;
	float deltaAngle = targetAngles[targetCount] - targetAngles[0];
	deltaAngle = Degree::normalizeAngle(deltaAngle);

	Vector tempVector;
	tempVector.setAsPolar(cautionLength, deltaAngle);

	angleWeight += ((180 - fabs(deltaAngle)) / 30.00) * .15;
	 
	targetWeights[targetCount] += angleWeight * angleWeightRate;
}

void DribblePath::setCraudWeight(int targetCount)
{
	float beginAngle = targetAngles[targetCount] - 90;
	beginAngle = Degree::normalizeAngle(beginAngle);
	Vector tempVector;
	
	float currentAngle;
	float craudWeight = 0;
	float playerDist;
	float negWeight = 0;
	float localNegWeight;
	Vector playerVector;
	
	for (int j = 0; j < FULL_PLAYERS_NUM; j++)
		if (worldModel->getFullPlayer(TID_OPPONENT, j).isValid())
		{
			playerVector.setByPoints(beginPoint, 
				worldModel->getFullPlayer(TID_OPPONENT, j).getPos());
			
			currentAngle = playerVector.getDirection() - beginAngle;	
			currentAngle = Degree::normalizeAngle(currentAngle);

			playerDist = worldModel->getFullPlayer(TID_OPPONENT, j).getDistance(beginPoint);			
			tempVector.setAsPolar(playerDist, currentAngle);

			if (playerDist < cautionLength * 1.2)
				isOpponent = 1;

			if (tempVector.getY() <= cautionLength && tempVector.getY() > 0 && 
					fabs(tempVector.getX()) <= (cautionLength * xRate))
			{
				localNegWeight = -(tempVector.getY() - fabs(tempVector.getX()));
				localNegWeight *= (fabs(tempVector.getX()) / cautionLength) / xRate;

				if (playerDist < 5)
				{
					float rate = 5 - playerDist;
					if (rate < 1.5)
						rate = 1.5;
	
					localNegWeight *= rate;
				}

				if (localNegWeight > 0)
					localNegWeight = 0;

				if (localNegWeight < negWeight)
					negWeight = localNegWeight;
			}
		}	

	for (int j = 0; j < HALF_PLAYERS_NUM; j++)
		if (worldModel->getHalfPlayer(TID_OPPONENT, j).isValid())
		{
			playerVector.setByPoints(beginPoint, 
				worldModel->getHalfPlayer(TID_OPPONENT, j).getPos());
			
			currentAngle = playerVector.getDirection() - beginAngle;	
			currentAngle = Degree::normalizeAngle(currentAngle);

			playerDist = worldModel->getHalfPlayer(TID_OPPONENT, j).getDistance(beginPoint);			
			tempVector.setAsPolar(playerDist, currentAngle);

			if (playerDist < cautionLength * 1.2)
				isOpponent = 1;

			if (tempVector.getY() <= cautionLength && tempVector.getY() > 0 && 
					fabs(tempVector.getX()) <= (cautionLength * xRate))
			{
				localNegWeight = -(tempVector.getY() - fabs(tempVector.getX()));
				localNegWeight *= (fabs(tempVector.getX()) / cautionLength) / xRate;

				if (playerDist < 5)
				{
					float rate = 5 - playerDist;
					if (rate < 1.5)
						rate = 1.5;
	
					localNegWeight *= rate;
				}

				if (localNegWeight > 0)
					localNegWeight = 0;

				if (localNegWeight < negWeight)
					negWeight = localNegWeight;
			}
		}	



	for (int j = 0; j < QUARTER_PLAYERS_NUM; j++)
		if (worldModel->getQuarterPlayer(j).isAlive())
		{
			playerVector.setByPoints(beginPoint, 
				worldModel->getQuarterPlayer(j).getPos());
			currentAngle = playerVector.getDirection() - beginAngle;	
			currentAngle = Degree::normalizeAngle(currentAngle);
	
			playerDist = worldModel->getQuarterPlayer(j).getDistance(beginPoint);			
			tempVector.setAsPolar(playerDist, currentAngle);
	
			if (playerDist < cautionLength * 1.2)
				isOpponent = 1;
	
			if (tempVector.getY() <= cautionLength && tempVector.getY() > 0 && 
					fabs(tempVector.getX()) <= (cautionLength * xRate))
			{
				localNegWeight = -(tempVector.getY() - fabs(tempVector.getX()));
				localNegWeight *= (fabs(tempVector.getX()) / cautionLength) / xRate;

				if (playerDist < 5)
				{
					float rate = 5 - playerDist;
					if (rate < 1.5)
						rate = 1.5;

					localNegWeight *= rate;
				}

				if (localNegWeight > 0)
					localNegWeight = 0;

				if (localNegWeight < negWeight)
					negWeight = localNegWeight;
			}
		}
		
	craudWeight = negWeight;
	targetWeights[targetCount] += craudWeight * craudWeightRate;	
}

void DribblePath::setTargetWeight(int targetCount)
{
	setCraudWeight(targetCount);
	setAngleWeight(targetCount);
//	LOG << "TotalWeight : " << targetAngles[targetCount] << " " << 
//		targetWeights[targetCount] << endl; 
}


void DribblePath::setDribbleTarget()
{
	Vector tempVector;
	
	tempVector.setByPoints(Point(beginPoint.x, 
			beginPoint.y), dribbleTarget);
	
	float beginAngle = tempVector.getDirection();   
	
	for (int i = 0; i < targetNums; i++) 
	{
		targetAngles[i] = beginAngle + (i * 10); 	
		targetAngles[i] = Degree::normalizeAngle(targetAngles[i]);
		setTargetWeight(i);
	}

	if (beginPoint.y >= 31)
		for (int i = 0; i < targetNums; i++)
			if (targetAngles[i] > 0 || (beginPoint.y >= 
				32 && targetAngles[i] >= 0))
				targetWeights[i] -= 1000;

	if (beginPoint.y <= -31)
		for (int i = 0; i < targetNums; i++)
			if (targetAngles[i] < 0 || (beginPoint.y <= 
				-32 && targetAngles[i] <= 0))
				targetWeights[i] -= 1000;
	
	if (worldModel->getBall().getPos().getX() >= 50)
		for (int i = 0; i < targetNums; i++)
			if (fabs(targetAngles[i]) < 90 || 
				(worldModel->getBall().getPos().getX() >= 51 && 
				fabs(targetAngles[i]) <= 90))
				targetWeights[i] -= 1000;

	bestDribbleTarget = 0;

	for (int i = 0; i < targetNums; i++)
		if (targetWeights[i] > targetWeights[bestDribbleTarget])
			bestDribbleTarget = i;

	Vector lastVector;
	lastVector.setAsPolar(10, targetAngles[bestDribbleTarget]);

	dribbleTarget.x = beginPoint.x + lastVector.getX();
	dribbleTarget.y = beginPoint.y + lastVector.getY();

//	LOG << "MainDribbletarget : " << dribbleTarget << endl;
}


void DribblePath::setDefaultParams()
{
	isOpponent = 0;	

	float firstXPathStep = 42;

	dribbleTarget.x = 52.5;

	if (worldModel->getBall().getPos().getX() < firstXPathStep)
		dribbleTarget.y = worldModel->getBall().getPos().getY();
	else
		dribbleTarget.y = 0;	

	targetNums = 36;
	cautionLength = 12;
	angleWeightRate = 1.6;
	craudWeightRate = 8;
	xRate = .25;
}

void DribblePath::setPathParams(int targetNumsArg, float cautionLengthArg,
					float angleWeightRateArg, float craudWeightRateArg,
					float xRateArg)
{
	float firstXPathStep = 42;

	dribbleTarget.x = 52.5;

	if (worldModel->getBall().getPos().getX() < firstXPathStep)
		dribbleTarget.y = worldModel->getBall().getPos().getY();
	else
		dribbleTarget.y = 0;	
	isOpponent = 0;

	targetNums = targetNumsArg;
	cautionLength = cautionLengthArg;
	angleWeightRate = angleWeightRateArg;
	craudWeightRate = craudWeightRateArg;
	xRate = xRateArg;
}					

void DribblePath::setBeginPoint(Point beginPointArg)
{
	beginPoint = beginPointArg;
}

float DribblePath::getBestTargetWeight()
{
	setDribbleTarget();
	return targetWeights[bestDribbleTarget];
}

float DribblePath::getTargetWeight(int targetCount)
{
	setDribbleTarget();
	return targetWeights[targetCount];
}

float DribblePath::getBestTargetDir()
{
	return targetAngles[bestDribbleTarget];
}

