/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Meisam Vosoughpour
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>

#include <BasicDash.h>
#include <Line.h>
#include <Vector.h>
#include <Degree.h>
#include <Logger.h>
#include <Command.h>
#include <BasicTurn.h>
#include <GoalieIntercept.h>

#include <AdvancedAgent.h>


using namespace std;

// class GoalieCatch
GoalieCatch::GoalieCatch(const WorldModel *worldModel):
		AdvancedAction(worldModel)
{
}

void GoalieCatch::execute(Form &form, const Library &library)
{
	if (catchableFlag == true)
		command = new CatchCommand(AT_CATCH,
				worldModel->getBall().getBodyVec().getDirection());
	else
		command = new KickCommand(AT_GOALIE_INTERCEPT, 100, 0);
		
}

float GoalieCatch::getValue(const Library &library)
{
	if (worldModel->getBall().getPos().getX() < -36.4 &&
		worldModel->getBody().getPos().getX() < -36.4 &&
		fabs(worldModel->getBall().getPos().getY()) < 19.6 &&
		fabs(worldModel->getBody().getPos().getY()) < 19.6)
	{
		if (library.lastBodyCycleCommand->getCreator() == AT_PASS)
		{
			catchableFlag = false;
			return AD_MIN_VALUE;
		}
		if (worldModel->isBallCatchable())
		{
			catchableFlag = true;
			return AD_ALWAYS_RUN_VALUE;
		}	
	}
/*	else if (worldModel->isBallKickable())
	{
		catchableFlag = false;	
		return AD_ALWAYS_RUN_VALUE;
	}*/
	// I do not want to do any action.
	catchableFlag = false;	
	return AD_MIN_VALUE;
}

// class GoalieIntercept
GoalieIntercept::GoalieIntercept(const WorldModel *worldModel): AdvancedAction(worldModel)
{
}

float GoalieIntercept::getValue(const Library &library)
{
	LOG << "GoalieIntercept::getValue" << endl;

	Body simBody;
	Ball simBall;
	Body waitBody;
	Ball waitBall;
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	DashCommand tempDashCommand(AT_GOALIE_INTERCEPT, 0);
	float bodyCatchableArea;
	float extraCatchableForNoise;
	float extraCatchableForAngleNoise = 0;
	int catchedBallCycles;
	bool catchedBall;
	int mainGoToBallTime = 0;
	float dashPower;
	float mainPositiveDashPower, mainNegativeDashPower;
	
	mainPositiveDashPower = 100;
	mainNegativeDashPower = -100;
	if (worldModel->getBody().getStamina() < 1400)
	{
		mainPositiveDashPower = worldModel->getBody().getStaminaIncMax();
		mainNegativeDashPower = -(worldModel->getBody().getStaminaIncMax() / 2);
	}
	
	bodyCatchableArea = hypot(worldModel->getBody().getCatchableAreaL(), 
						worldModel->getBody().getCatchableAreaW() / 2.00); 

	gtbPointsNumber = 0;
	for (int i = 0; i < GOALIE_MAX_GTB_POINTS_NUM; i++)
	{
		gtbTime[i] = 999; 
		minBallDistance[i] = bodyCatchableArea;
		extraGTBTime[i] = 0;
		gtbPointWeight[i] = 0;
		gtbCommand[i] = NULL;//TurnToPoint(Point(50, 0), worldModel->getBody()).getCommand();
	}
	
////////NoAction
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	waitBody = simBody;
	waitBall = simBall;
	
	

	for (int i = 0; i <= GOALIE_WAIT_FOR_BALL_CYCLES; i++)
	{
//		float ExtraCatchableForNoise = .15 * sqrt(MainGoToBallTime);
		extraCatchableForNoise = -.15;
		if (waitBall.getBodyVec().getMagnitude() <= (bodyCatchableArea + extraCatchableForNoise))
		{
				gtbPoint[gtbPointsNumber] = Point(waitBody.getPos().getX(), waitBody.getPos().getY()); 
			extraGTBTime[gtbPointsNumber] = i;
			gtbTime[gtbPointsNumber] = mainGoToBallTime + (extraGTBTime[gtbPointsNumber] / 2.00);
//*tof*/			gtbTime[gtbPointsNumber] -= 2;				
			minBallDistance[gtbPointsNumber] = waitBall.getBodyVec().getMagnitude();
/*TOF2*/	gtbPointWeight[gtbPointsNumber] += min((bodyCatchableArea - minBallDistance[gtbPointsNumber]) , (float).3) * 10; 		
			distToBegin[gtbPointsNumber] = hypot(waitBall.getPos().getX() - worldModel->getBall().getPos().getX(), waitBall.getPos().getY() - worldModel->getBall().getPos().getY());
			gtbPointWeight[gtbPointsNumber] -= 1000;
			catchedBall = 1;
			gtbPointsNumber++;
		}
		waitBody.simulateByDynamics();
		waitBall.simulateByDynamics(waitBody);
	}
	
//	LOG << "MeisamGoToBall : gtbPointsNumberT& : " << gtbPointsNumber << endl;
////////positiveDash	
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	catchedBallCycles = 0;
	catchedBall = 0;
	mainGoToBallTime = 0;
	dashPower = 100;
	if (simBody.getStamina() < 1400)
				dashPower = simBody.getStaminaIncMax();
	tempDashCommand.setPower(dashPower);
	waitBody = simBody;
	waitBall = simBall;
	while (catchedBallCycles <= 3 && mainGoToBallTime < GOALIE_MAX_POSITIVE_DASH_CYCLES)
	{
		mainGoToBallTime++;
		
//KOMAK		simBody.setStamina(simBody.getStamina() + simBody.getStaminaIncMax());
		waitBody = simBody;
		waitBall = simBall;
		
		dashPower = 100;
		if (simBody.getStamina() < 1400)
			dashPower = simBody.getStaminaIncMax();
		tempDashCommand.setPower(dashPower);
		delete gtbCommand[gtbPointsNumber];
		gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, dashPower);
		simBody.simulateByAction(dynamic_cast<Command *> (&tempDashCommand));
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
//KOMAK		simBody.setStamina(simBody.getStamina() - DashPower);

		
		for (int i = 0; i <= GOALIE_WAIT_FOR_BALL_CYCLES; i++)
		{
			extraCatchableForNoise = worldModel->getBall().getBodyVec().getMagnitude() * .12;
			extraCatchableForNoise += .3;
/////////////////////POSITIVE ANGLE NOISE			
			Line bodyAngleLine, ballVelLine;
			Point interceptGTBPoint;
			bodyAngleLine.setBySourceDir(Point(worldModel->getBody().getPos().getX(), worldModel->getBody().getPos().getY()), worldModel->getBody().getBodyDir());
			ballVelLine.setBySourceDir(Point(worldModel->getBall().getPos().getX(), worldModel->getBall().getPos().getY()), worldModel->getBody().getVel().getDirection());
			bodyAngleLine.getLineIntersect(ballVelLine, interceptGTBPoint);
			Vector interceptGTBVector;
			interceptGTBVector.setByPoints(Point(worldModel->getBody().getPos().getX(), worldModel->getBody().getPos().getY()), interceptGTBPoint);
			if (fabs(Degree::normalizeAngle(interceptGTBVector.getDirection() - worldModel->getBody().getBodyDir())) < 2)
			{
				float deltaAngleToBallAngle;
				deltaAngleToBallAngle = simBody.getBodyDir() - simBall.getVel().getDirection();
				deltaAngleToBallAngle = Degree::normalizeAngle(deltaAngleToBallAngle);
				if (interceptGTBVector.getMagnitude() < 8)
				{
					if (fabs(deltaAngleToBallAngle) < 90)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if (fabs(deltaAngleToBallAngle) < 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * .2;
					else if (fabs(deltaAngleToBallAngle) >= 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * .4;
				}
				else
				{
					if (fabs(deltaAngleToBallAngle) < 90)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if (fabs(deltaAngleToBallAngle) < 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if (fabs(deltaAngleToBallAngle) >= 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
				}
			}
			else
			{
				extraCatchableForAngleNoise = 0;
				extraCatchableForNoise /= 2;
			}

			if (waitBall.getBodyVec().getMagnitude() <= (bodyCatchableArea + extraCatchableForNoise + extraCatchableForAngleNoise))
			{
				gtbPoint[gtbPointsNumber] = Point(waitBody.getPos().getX(), waitBody.getPos().getY()); 
				extraGTBTime[gtbPointsNumber] = i;
				gtbTime[gtbPointsNumber] = mainGoToBallTime + (extraGTBTime[gtbPointsNumber] / 2.00);
///*TOF*/				gtbTime[gtbPointsNumber] -= 0;				
				minBallDistance[gtbPointsNumber] = waitBall.getBodyVec().getMagnitude();
/*TOF2*/	gtbPointWeight[gtbPointsNumber] += min((bodyCatchableArea - minBallDistance[gtbPointsNumber]) , (float).3) * 10; 		
				distToBegin[gtbPointsNumber] = hypot(waitBall.getPos().getX() - worldModel->getBall().getPos().getX(), waitBall.getPos().getY() - worldModel->getBall().getPos().getY());
				delete gtbCommand[gtbPointsNumber];
				gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, mainPositiveDashPower);
				catchedBall = 1;
				gtbPointsNumber++;
			}
			waitBody.simulateByDynamics();
			waitBall.simulateByDynamics(waitBody);
		}
		
		if (catchedBall) 
				catchedBallCycles++;
	}
//	LOG << "MeisamGoToBall : gtbPointsNumberT? : " << gtbPointsNumber << endl;
	
//////////////////Negative Dash	

	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	catchedBallCycles = 0;
	catchedBall = 0;
	mainGoToBallTime = 0;
	dashPower = -100;
	if (simBody.getStamina() < 1400)
		dashPower = -simBody.getStaminaIncMax() / 2.00;
	tempDashCommand.setPower(-dashPower);
	while (catchedBallCycles <= 3 && mainGoToBallTime < GOALIE_MAX_NEGATIVE_DASH_CYCLES)
	{
		mainGoToBallTime++;
		
//KOMAK		simBody.setStamina(simBody.getStamina() + simBody.getStaminaIncMax());
		waitBody = simBody;
		waitBall = simBall;
		
		dashPower = -100;
		if(simBody.getStamina() < 1400)
			dashPower = -simBody.getStaminaIncMax() / 2.00;
		tempDashCommand.setPower(dashPower);
		delete gtbCommand[gtbPointsNumber];
		gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, dashPower);
		simBody.simulateByAction(&tempDashCommand);
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
//KOMAK		simBody.setStamina(simBody.getStamina() + dashPower * 2);////dar inja dashpower manfi ast.
		
		for (int i = 0; i <= GOALIE_WAIT_FOR_BALL_CYCLES; i++)
		{
			extraCatchableForNoise = worldModel->getBall().getBodyVec().getMagnitude() * .12;
			extraCatchableForNoise += .3;
			float deltaAngleToBallAngle, negBodyAngle;
			negBodyAngle = simBody.getBodyDir() + 180;
			deltaAngleToBallAngle = negBodyAngle - simBall.getVel().getDirection();
			deltaAngleToBallAngle = Degree::normalizeAngle(deltaAngleToBallAngle);

/////////////////////NEGATIVE ANGLE NOISE			
			Line bodyAngleLine, ballVelLine;
			Point interceptGTBPoint;
			bodyAngleLine.setBySourceDir(Point(simBody.getPos().getX(), simBody.getPos().getY()), negBodyAngle);			
			ballVelLine.setBySourceDir(Point(simBall.getPos().getX(), simBall.getPos().getY()), worldModel->getBall().getVel().getDirection());
			bodyAngleLine.getLineIntersect(ballVelLine, interceptGTBPoint);
			Vector interceptGTBVector;
			interceptGTBVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), interceptGTBPoint);
			if (fabs(Degree::normalizeAngle(interceptGTBVector.getDirection() - negBodyAngle)) < 2)
			{
				if (interceptGTBVector.getMagnitude() < 8)
				{
					if (fabs(deltaAngleToBallAngle) < 90)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if (fabs(deltaAngleToBallAngle) < 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * .2;
					else if (fabs(deltaAngleToBallAngle) >= 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * .4;
				}
				else
				{
					if (fabs(deltaAngleToBallAngle) < 90)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if (fabs(deltaAngleToBallAngle) < 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
					else if(fabs(deltaAngleToBallAngle) >= 120)
						extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.5;
				}
			}
			else
			{
				extraCatchableForAngleNoise = 0;
				extraCatchableForNoise /= 2.00;
			}		
/////////////////////NEGATIVE ANGLE NOISE			
			if (waitBall.getBodyVec().getMagnitude() <= (bodyCatchableArea + extraCatchableForNoise + extraCatchableForAngleNoise))
			{
				gtbPoint[gtbPointsNumber] = Point(waitBody.getPos().getX(), waitBody.getPos().getY()); 
				extraGTBTime[gtbPointsNumber] = i;
				gtbTime[gtbPointsNumber] = mainGoToBallTime + (extraGTBTime[gtbPointsNumber] / 2.00);
//				gtbTime[gtbPointsNumber] -= 2;				
				minBallDistance[gtbPointsNumber] = waitBall.getBodyVec().getMagnitude();
/*TOF2*/	gtbPointWeight[gtbPointsNumber] += min((bodyCatchableArea - minBallDistance[gtbPointsNumber]) , (float).3) * 10; 		
				distToBegin[gtbPointsNumber] = hypot(waitBall.getPos().getX() - worldModel->getBall().getPos().getX(), waitBall.getPos().getY() - worldModel->getBall().getPos().getY());
				delete gtbCommand[gtbPointsNumber];
				gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, mainNegativeDashPower);
				catchedBall = 1;
				gtbPointsNumber++;
			}
			waitBody.simulateByDynamics();
			waitBall.simulateByDynamics(waitBody);
		}
		
		if (catchedBall) 
				catchedBallCycles++;
	}
//	LOG << "MeisamGoToBall : gtbPointsNumberT! : " << gtbPointsNumber << endl;
///////////////OneTurnPositiveDash
	
	Body virtualDashBody;
	virtualDashBody = simBody;
	virtualDashBody.setVel().setAsCartesian(0,0);
	Vector goToBallVector;
	float deltaAngleToGTB;
	float gtbRadius = bodyCatchableArea;
//	LogMaker.Log("GTBRadius 1 : ", GTBRadius); 
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	catchedBallCycles = 0;
	catchedBall = 0;
	mainGoToBallTime = 0;
	extraCatchableForNoise = -.045 * mainGoToBallTime;
	simBody.simulateByDynamics();
	simBall.simulateByDynamics(simBody);
	goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
	deltaAngleToGTB = simBody.getBodyDir() - goToBallVector.getDirection();
	deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);
//	LogMaker.Log("DeltaAngleToGTB 1 : ", DeltaAngleToGTB);
	mainGoToBallTime++;
//KOMAK	virtualDashBody.setStamina(virtualDashBody.setStamina() + virtualDashBody.getStaminaIncMax() * 1);/*One Turn*/;
	dashPower = 100;
	if (virtualDashBody.getStamina() < 1400)
		dashPower = virtualDashBody.getStaminaIncMax();
	tempDashCommand.setPower(dashPower);	
	while (mainGoToBallTime < GOALIE_MAX_GTB_CHECKED_TIME && gtbPointsNumber < 30)
	{
		mainGoToBallTime++;
//KOMAK		virtualDashBody.setStamina(virtualDashBody.getStamina() + virtualDashBody.getStaminaIncMax() * 1);/*One Turn*/;
		
		dashPower = 100;
		if (virtualDashBody.getStamina() < 1400)
				dashPower = virtualDashBody.getStaminaIncMax();
		tempDashCommand.setPower(dashPower);	
		virtualDashBody.simulateByAction(&tempDashCommand);
//KOMAK		virtualDashBody.setStamina(virtualDashBody.getStamina() - dashPower);
		gtbRadius += virtualDashBody.getVel().getMagnitude();
		
		extraCatchableForNoise = - .045 * mainGoToBallTime;
		extraCatchableForNoise -= .45;
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
		goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		deltaAngleToGTB = goToBallVector.getDirection() - simBody.getBodyDir();
		deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);
		float deltaAngleToBallAngle;
		Vector gtbVector;
		gtbVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		deltaAngleToBallAngle = gtbVector.getDirection() - simBall.getVel().getDirection();
		deltaAngleToBallAngle = Degree::normalizeAngle(deltaAngleToBallAngle);

		if (goToBallVector.getMagnitude() < 10)
		{
			if (fabs(deltaAngleToBallAngle) < 90)
//				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			extraCatchableForAngleNoise = 0;
			
			else if (fabs(deltaAngleToBallAngle) < 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			else if (fabs(deltaAngleToBallAngle) >= 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
		}
		else
		{
				if (fabs(deltaAngleToBallAngle) < 90)
//				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			extraCatchableForAngleNoise = 0;
			else if (fabs(deltaAngleToBallAngle) < 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.8;
			else if (fabs(deltaAngleToBallAngle) >= 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.8;
		}

		if ((gtbRadius + extraCatchableForNoise  + extraCatchableForAngleNoise) >= goToBallVector.getMagnitude() && fabs(deltaAngleToGTB) <= BasicTurn::getTurnAngle(179,worldModel->getBody())) 
		{
			gtbPoint[gtbPointsNumber] = Point(simBody.getPos().getX(), simBody.getPos().getY()); 
			gtbTime[gtbPointsNumber] = mainGoToBallTime;
			extraGTBTime[gtbPointsNumber] = 0;
			minBallDistance[gtbPointsNumber] = goToBallVector.getMagnitude() - gtbRadius + bodyCatchableArea;
			distToBegin[gtbPointsNumber] = hypot(simBall.getPos().getX() - worldModel->getBall().getPos().getX(), simBall.getPos().getY() - worldModel->getBall().getPos().getY());
				delete gtbCommand[gtbPointsNumber]; 
			gtbCommand[gtbPointsNumber] = TurnAngle(AT_GOALIE_INTERCEPT, deltaAngleToGTB, worldModel->getBody()).getCommand();
/*TOF*/			if (fabs(deltaAngleToGTB) <= 8)
{
				delete gtbCommand[gtbPointsNumber]; 
/*TOF*/				gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, mainPositiveDashPower);
}
			gtbPointsNumber++;
		}	
	}
	
//	LOG << "MeisamGoToBall : gtbPointsNumber 2 : " << gtbPointsNumber << endl;


///////////////OneTurnNegativeDash
	
/*	virtualDashBody = simBody;
	virtualDashBody.setVel().setAsCartesian(0,0);
	gtbRadius = bodyCatchableArea;
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	catchedBallCycles = 0;
	catchedBall = 0;
	mainGoToBallTime = 0;
	extraCatchableForNoise = .06 * mainGoToBallTime;
	simBody.simulateByDynamics();
	simBall.simulateByDynamics(simBody);
	goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
	deltaAngleToGTB = simBody.getBodyDir() - goToBallVector.getDirection();
	deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);
	mainGoToBallTime++;
//KOMAK	virtualDashBody.setStamina(virtualDashBody.getStamina() + virtualDashBody.getStaminaIncMax() * 1);
	dashPower = -100;
	if (virtualDashBody.getStamina() < 1400)
		dashPower = -virtualDashBody.getStaminaIncMax() / 2.00;
   tempDashCommand.setPower(dashPower);	
	while (mainGoToBallTime < GOALIE_MAX_GTB_CHECKED_TIME && gtbPointsNumber < 50)
	{
		mainGoToBallTime++;
//KOMAK		virtualDashBody.setStamina(virtualDashBody.getStamina() + virtualDashBody.getStaminaIncMax() * 1);
		
		dashPower = -100;
		if (virtualDashBody.getStamina() < 1400)
				dashPower = -virtualDashBody.getStaminaIncMax() / 2.00;
		tempDashCommand = DashCommand(dashPower);	
		virtualDashBody.simulateByAction(&dashCommand);
//KOMAK				virtualDashBody.setStamina(virtualDashBody.getStamina() + DashPower * 2);
		gtbRadius += virtualDashBody.getVel().getMagnitude();
		
		extraCatchableForNoise = .06 * mainGoToBallTime;
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
		goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		goToBallVector.setAsPolar(goToBallVector.getMagnitude(), Degree::normalizeAngle(goToBallVector.getDirection() + 180));
		deltaAngleToGTB = goToBallVector.getMagnitude() - simBody.getBodyDir();
		deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);

//KOMAK		if((gtbRadius + extraCatchableForNoise) >= goToBallVector.getMagnitude() && fabs(deltaAngleToGTB) <= calculate.Public.GetVirtualTurnAngle(179)) 
		{
			gtbPoint[gtbPointsNumber] = Point(simBody.getPos().getX(), simBody.getPos().getY()); 
			gtbTime[gtbPointsNumber] = mainGoToBallTime;
			extraGTBTime[gtbPointsNumber] = 0;
			minBallDistance[gtbPointsNumber] = goToBallVector.getMagnitude() - gtbRadius + bodyCatchableArea;
			distToBegin[gtbPointsNumber] = hypot(simBall.getPos().getX() - worldModel->getBall().getPos().getX(), simBall.getPos().getY() - worldModel->getBall().getPos().getY());
			gtbCommand[gtbPointsNumber] = *TurnAngle(deltaAngleToGTB, worldModel->getBody());
			gtbPointsNumber++;
		}	
	}*/
//	LOG << "MeisamGoToBall : gtbPointsNumber 3 : " << gtbPointsNumber << endl;

///////////////TwoTurnPositiveDash
	
	virtualDashBody = simBody;
	virtualDashBody.setVel().setAsCartesian(0,0);
	gtbRadius = bodyCatchableArea;
	simBody = worldModel->getBody();
	simBall = worldModel->getBall();
	catchedBallCycles = 0;
	catchedBall = 0;
	mainGoToBallTime = 0;
	extraCatchableForNoise = - .045 * mainGoToBallTime;
	for (int i = 0; i < 2; i++)
	{
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
		goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		deltaAngleToGTB = simBody.getBodyDir() - goToBallVector.getDirection();
		deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);
		mainGoToBallTime++;
//KOMAK				virtualDashBody.setStamina(virtualDashBody.getStamina() + virtualDashBody.getStaminaIncMax());
	}
	dashPower = 100;
	if (virtualDashBody.getStamina() < 1400)
		dashPower = virtualDashBody.getStaminaIncMax();
	tempDashCommand.setPower(dashPower);	
	while (mainGoToBallTime < GOALIE_MAX_GTB_CHECKED_TIME && gtbPointsNumber < 70)
	{
		mainGoToBallTime++;
//		virtualDashBody.setStamina()(virtualDashBody.getStamina() + virtualDashBody.getStaminaIncMax());
		
		dashPower = 100;
		if (virtualDashBody.getStamina() < 1400)
				dashPower = virtualDashBody.getStaminaIncMax();
		tempDashCommand.setPower(dashPower);	
		virtualDashBody.simulateByAction(dynamic_cast <Command *> (&tempDashCommand));
//KOMAK				virtualDashBody.setStamina(virtualDashBody.getStamina() - dashPower);
		gtbRadius += virtualDashBody.getVel().getMagnitude();
		
		extraCatchableForNoise = -.045 * mainGoToBallTime;
		extraCatchableForNoise -= .45;
		simBody.simulateByDynamics();
		simBall.simulateByDynamics(simBody);
		goToBallVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		deltaAngleToGTB = goToBallVector.getDirection() - simBody.getBodyDir();
		deltaAngleToGTB = Degree::normalizeAngle(deltaAngleToGTB);
		float deltaAngleToBallAngle;
		Vector gtbVector;
		gtbVector.setByPoints(Point(simBody.getPos().getX(), simBody.getPos().getY()), Point(simBall.getPos().getX(), simBall.getPos().getY()));
		deltaAngleToBallAngle = gtbVector.getDirection() - simBall.getVel().getDirection();
		deltaAngleToBallAngle = Degree::normalizeAngle(deltaAngleToBallAngle);
		if (goToBallVector.getMagnitude() < 10)
		{
			if (fabs(deltaAngleToBallAngle) < 90)
//				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			extraCatchableForAngleNoise = 0;
			else if (fabs(deltaAngleToBallAngle) < 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			else if (fabs(deltaAngleToBallAngle) >= 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
		}
		else
		{
				if (fabs(deltaAngleToBallAngle) < 90)
//				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1;
			extraCatchableForAngleNoise = 0;
			else if (fabs(deltaAngleToBallAngle) < 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.8;
			else if (fabs(deltaAngleToBallAngle) >= 120)
				extraCatchableForAngleNoise = (Degree::cos(fabs(deltaAngleToBallAngle)) + 0) * 1.8;
		}

		if ((gtbRadius + extraCatchableForNoise + extraCatchableForAngleNoise) >= goToBallVector.getMagnitude())// && fabs(deltaAngleToGTB) <= Calculate.Public.GetVirtualTurnAngle(180)) 
		{
			gtbPoint[gtbPointsNumber] = Point(simBody.getPos().getX(), simBody.getPos().getY()); 
			gtbTime[gtbPointsNumber] = mainGoToBallTime;
			extraGTBTime[gtbPointsNumber] = 0;
			minBallDistance[gtbPointsNumber] = goToBallVector.getMagnitude() - gtbRadius + bodyCatchableArea;
			distToBegin[gtbPointsNumber] = hypot(simBall.getPos().getX() - worldModel->getBall().getPos().getX(), simBall.getPos().getY() - worldModel->getBall().getPos().getY());
				delete gtbCommand[gtbPointsNumber];
			gtbCommand[gtbPointsNumber] = TurnAngle(AT_GOALIE_INTERCEPT, deltaAngleToGTB, worldModel->getBody()).getCommand();
/*TOF*/			if (fabs(deltaAngleToGTB) <= 8)
{
				delete gtbCommand[gtbPointsNumber];
/*TOF*/				gtbCommand[gtbPointsNumber] = new DashCommand(AT_GOALIE_INTERCEPT, mainPositiveDashPower);
}
			gtbPointsNumber++;
		}	
	}

//	LOG << "MeisamGoToBall : gtbPointsNumber 4 : " << gtbPointsNumber << endl;
	for (int i = 0; i < gtbPointsNumber; i++)
	{
		gtbPointWeight[i] -= gtbTime[i];
//		gtbPointWeight[i] += extraGTBTime[i];	
	}
	
	bestGTBPointNumber = 0;
	for (int i = 0; i < gtbPointsNumber; i++)
	{
//		LOG << "i , gtbTime[i] : " <<  i << " " << gtbTime[i] << endl;
//		LOG << "i , gtbWeightPoint[i] : " << i << " " <<  gtbPointWeight[i] << endl;
//		LOG << "i , extraGTBTime[i] : " << i << " " << extraGTBTime[i] << endl;
//		LOG << "i , minBallDistance[i] : " << i << " " << minBallDistance[i] << endl;
		if (gtbPointWeight[i] > gtbPointWeight[bestGTBPointNumber])
			bestGTBPointNumber = i; 
	}
	
	if (dynamic_cast <DashCommand *> (gtbCommand[bestGTBPointNumber]) != NULL && 
		worldModel->getBall().getVel().getMagnitude() > .4 && 
		worldModel->isOppBallShooted(worldModel->getBall()))
	{

		float deltaDirToBallVel;
		deltaDirToBallVel = worldModel->getBall().getVel().getDirection() - 
							worldModel->getBody().getBodyDir();
		deltaDirToBallVel = Degree::normalizeAngle(deltaDirToBallVel);
		deltaDirToBallVel = fabs(deltaDirToBallVel);
		if (deltaDirToBallVel < 170 && deltaDirToBallVel > 10)
		{
			Line bodyLine, ballLine;
			Point interPoint;
			ballLine.setBySourceDir(worldModel->getBall().getPos(), 
					worldModel->getBall().getVel().getDirection());
			bodyLine.setBySourceDir(worldModel->getBody().getPos(), 
					worldModel->getBody().getBodyDir());
			ballLine.getLineIntersect(bodyLine, interPoint);
			delete gtbCommand[bestGTBPointNumber];
			gtbCommand[bestGTBPointNumber] = PreciseGotoPoint(AT_GOALIE_INTERCEPT, interPoint, .2,
						worldModel->getBody(), 1).getCommand();
		}
	} 
	if (gtbPoint[bestGTBPointNumber].x > -35.5 ||
		fabs(gtbPoint[bestGTBPointNumber].y) > 19.5)
		return AD_MIN_VALUE;
	else
		return AD_ALWAYS_RUN_VALUE;	
//		LOG << "BestGTBCommand : " << gtbCommand[bestGTBPointNumber]->toString() << endl;		
//	LOG << "BestGTBPointNumber : " << bestGTBPointNumber << endl;

	return AD_NORMAL_VALUE;
}

void GoalieIntercept::execute(Form &form, const Library &library)
{
	LOG << "GoalieIntercept::execute" << endl;

	command = gtbCommand[bestGTBPointNumber];
}

