/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Meisam Vosoughpour
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>

#include <Line.h>
#include <Degree.h>
#include <Vector.h>
#include <Logger.h>
#include <BasicTurn.h>
#include <BasicDash.h>
#include <Basics.h>
#include <GoaliePositioning.h>


using namespace std;

GoaliePositioning::GoaliePositioning(const WorldModel *worldModel):
		AdvancedAction(worldModel)
{
}

void GoaliePositioning::setGoalieParams()
{
    Point ballPoint;
    float changeXDist = 1;
    float changeYDistBadAngle = 1;
    float minGoalieX = -48.5;
    float virtualMinGoalieX;
//  	float overLoadDistByNoTurn;
// 		float overLoadDistByTurn;
    float virtualGoalieX, virtualGoalieYIn52, 
		virtualGoalieYIn44, virtualGoalieYInNow;
    
	Line midAngleLine, upWallLine, downWallLine, directWallLine; 
    //KOMAK
		/*KOMAK*/ float goalieXHome = -47;
	//KOMAK
	if (worldModel->getPlayMode() == PM_PENALTY_SETUP ||
		worldModel->getPlayMode() == PM_PENALTY_SETUP_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY)
	{
		goalieXHome = -40;
		changeXDist = 2;
	}	
	else
	{
//	 LATEST TRICK	
	if (config["Agent"]["Public"]["LatestTrick"].asBool() && 
		worldModel->getBall().getPos().getX() > -3)
	{
		if (worldModel->getBall().getPos().getX() > -34 &&
			worldModel->getBall().getPos().getX() < -15)
			goalieXHome = Basics::reRate(worldModel->getBall().getPos().getX(),
					-25, -15, -47, -38);
		
		if (worldModel->getBall().getPos().getX() > -15 &&
			worldModel->getBall().getPos().getX() < 0)
			goalieXHome = Basics::reRate(worldModel->getBall().getPos().getX(),
					-15, 0, -38, -20);

	
		if (worldModel->getBall().getPos().getX() >= 0)
			goalieXHome = Basics::reRate(worldModel->getBall().getPos().getX(),
					0, 54, -24, -15);
		if (goalieXHome > -32)
			goalieXHome = -32;
	}
//	LATEST TRICK 				
	}
	
    virtualGoalieX = goalieXHome;
    virtualGoalieYIn52 = 6;
    virtualGoalieYIn44 = 6;

	ballPoint.x = worldModel->getBall().getPos().getX() + 
					worldModel->getBall().getVel().getX();
					
	ballPoint.y = worldModel->getBall().getPos().getY() + 
					worldModel->getBall().getVel().getY();

    if(ballPoint.x < -52.4)
		ballPoint.x = -52.4;

    virtualMinGoalieX = -52.5 + float((ballPoint.x - (-52.5)) / 2.8);   
    
    if(worldModel->getBall().getPos().getX() < -25 && 
		worldModel->getBody().getPos().getX() < virtualGoalieX)
    {
		if (worldModel->getBody().getPos().getX() > virtualMinGoalieX)
	   		virtualGoalieX = worldModel->getBody().getPos().getX(); 
		else if (virtualMinGoalieX < virtualGoalieX) 
			virtualGoalieX = virtualMinGoalieX;
    }

    if(virtualGoalieX < minGoalieX)
		virtualGoalieX = minGoalieX;

    if(fabs(virtualGoalieX - worldModel->getBody().getPos().getX()) 
		< changeXDist)
		virtualGoalieX = worldModel->getBody().getPos().getX();
    
    virtualGoalieYInNow = virtualGoalieYIn44 + ((-44 - 
							virtualGoalieX) * (virtualGoalieYIn52 - 
							virtualGoalieYIn44)) / (-44 - (-52.5));	 

    Point tirUpPoint, tirDownPoint;
    Vector tirUpVector, tirDownVector, midAngleVector;
    float deltaAngle, betweenAngle;
    
    tirUpPoint.x = tirDownPoint.x = -52.5;
    tirUpPoint.y = 7;
    tirDownPoint.y = -7;
    
    tirUpVector.setByPoints(ballPoint, tirUpPoint);
    tirDownVector.setByPoints(ballPoint, tirDownPoint);
    
    deltaAngle = tirUpVector.getDirection() - tirDownVector.getDirection();
    deltaAngle = Degree::normalizeAngle(deltaAngle);

    deltaAngle = fabs(deltaAngle);
    deltaAngle /= 2.00;
    
    betweenAngle = tirUpVector.getDirection() + deltaAngle;
    betweenAngle = Degree::normalizeAngle(betweenAngle);
    
    midAngleVector.setAsPolar(1, betweenAngle);
    midAngleLine.setBySourceDir(ballPoint, midAngleVector.getDirection());
    
    upWallLine.setByPoints(Point(virtualGoalieX, virtualGoalieYInNow), 
							Point(-52.5, virtualGoalieYIn52));
						
    downWallLine.setByPoints(Point(virtualGoalieX, -virtualGoalieYInNow), 
							Point(-52.5, -virtualGoalieYIn52));
							
    directWallLine.setByPoints(Point(virtualGoalieX, -1), 
								Point(virtualGoalieX, 1));

	LOG << "directWallLine : a, b, c = " << directWallLine.getA() << "," << directWallLine.getB() << "," << directWallLine.getC() << endl;
	LOG << "midAngleLine : a, b, c = " << midAngleLine.getA() << "," << midAngleLine.getB() << "," << midAngleLine.getC() << endl;
	LOG << "midAngleLine : direction = " << midAngleLine.getDirection() << endl;
	LOG << "virtualGoalieX = " << virtualGoalieX << endl;

    midAngleLine.getLineIntersect(directWallLine, goalieHomePoint);

    LOG << "goalieHomePoint 1 : " << goalieHomePoint << endl;
    
    if(goalieHomePoint.y > virtualGoalieYInNow)
    {
		midAngleLine.getLineIntersect(upWallLine, goalieHomePoint);
    	LOG << "goalieHomePoint 2 : " << goalieHomePoint << endl;
    }
    else if(goalieHomePoint.y < (-virtualGoalieYInNow))
	{
	    midAngleLine.getLineIntersect(downWallLine, goalieHomePoint);
    	LOG << "goalieHomePoint 3 : " << goalieHomePoint << endl;
	}

    mainIdealBetweenAngle = midAngleVector.getDirection(); 
    
    Vector goalieHomeVector;
    goalieHomeVector.setByPoints(worldModel->getBody().getPos(), 
								goalieHomePoint);    
    
	float localDeltaAngle = goalieHomeVector.getDirection() - 
							worldModel->getBody().getBodyDir();
							
    localDeltaAngle = Degree::normalizeAngle(localDeltaAngle);
    localDeltaAngle = fabs(localDeltaAngle);
    
	if((localDeltaAngle >= 15 || localDeltaAngle >= 165) && 
		fabs(goalieHomePoint.y - worldModel->getBody().getPos().getY()) <= 
		changeYDistBadAngle)
		goalieHomePoint.y = worldModel->getBody().getPos().getY();

    LOG << "goalieHomePoint 4 : " << goalieHomePoint << endl;

    float idealBodyAngle = mainIdealBetweenAngle - 90;
    idealBodyAngle = Degree::normalizeAngle(idealBodyAngle);

    Line idealAngleLine, ballToHomeLine;
    Point bodyPoint;
	
    bodyPoint.x = worldModel->getBody().getPos().getX();
    bodyPoint.y = worldModel->getBody().getPos().getY();
    
	idealAngleLine.setBySourceDir(bodyPoint, idealBodyAngle);

    Line tirUpLine, tirDownLine;
    tirUpLine.setByPoints(ballPoint, tirUpPoint);
    tirDownLine.setByPoints(ballPoint, tirDownPoint);

    LOG << "goalieHomePoint 5 : " << goalieHomePoint << endl;

    float miniChangeDist = .3;
	
    if(worldModel->getBall().getBodyVec().getMagnitude() >= 10)
		miniChangeDist += (worldModel->getBall().getBodyVec().getMagnitude() 
							- 10) * .05;
    
	if(worldModel->getBall().getBodyVec().getMagnitude() >= 25)
		miniChangeDist = 1.2;
    
	if (worldModel->getPlayMode() != PM_PLAY_ON)
    	miniChangeDist = 5;    
    
	if(mainIdealBetweenAngle > -122 && mainIdealBetweenAngle < 0)
    {
		goalieHomePoint = shootToTirSimulate(tirUpPoint, goalieHomePoint, 
							idealBodyAngle, .00/*firstDashLength*/, 
							6/*maxChangeDistance*/,miniChangeDist);
							
    	LOG << "goalieHomePoint 6 : " << goalieHomePoint << endl;
    }
	
    if(mainIdealBetweenAngle < 122 && mainIdealBetweenAngle > 0)
    {
		goalieHomePoint = shootToTirSimulate(tirDownPoint, goalieHomePoint, 
							idealBodyAngle, .00/*firstDashLength*/, 
							6/*maxChangeDistance*/,miniChangeDist);
							
    	LOG << "goalieHomePoint 7 : " << goalieHomePoint << endl;
    }    

    if(hypot(goalieHomePoint.x - worldModel->getBody().getPos().getX(), 
		goalieHomePoint.y - worldModel->getBody().getPos().getY()) < 3)
    {
    	if(mainIdealBetweenAngle < -122 && mainIdealBetweenAngle > -160) 
    	{
			Point tempSource;
			idealAngleLine.getLineIntersect(tirUpLine, tempSource);
			goalieHomePoint = shootToTirSimulate(tempSource, goalieHomePoint, 
								idealBodyAngle, .00/*firstDashLength*/, 
								7/*maxChangeDistance*/, .3);
								
 		   	LOG << "goalieHomePoint 8 : " << goalieHomePoint << endl;
    	}
		
	    if(mainIdealBetweenAngle > 124 && mainIdealBetweenAngle < 160) 
	    {
			Point tempSource;
			idealAngleLine.getLineIntersect(tirDownLine, tempSource);
			goalieHomePoint = shootToTirSimulate(tempSource, goalieHomePoint, 
								idealBodyAngle, .00/*firstDashLength*/, 
								7/*maxChangeDistance*/, .3);
								
  		  	LOG << "goalieHomePoint 9 : " << goalieHomePoint << endl;
    	}
    }

    LOG << "goalieHomePoint 10 : " << goalieHomePoint << endl;
    
    ballToHomeLine.setByPoints(goalieHomePoint, ballPoint);

    float permitedDist = 1.2;
    
	if(hypot(goalieHomePoint.x - worldModel->getBody().getPos().getX(), 
		goalieHomePoint.y - worldModel->getBody().getPos().getY()) < permitedDist)
		idealAngleLine.getLineIntersect(ballToHomeLine, goalieHomePoint);

   	LOG << "goalieHomePoint 11 : " << goalieHomePoint << endl;

    if(goalieHomePoint.x < -50.5 )
		goalieHomePoint.x = -50.5;

   	LOG << "goalieHomePoint 12 : " << goalieHomePoint << endl;
}

void GoaliePositioning::setGoalieParamsLineMode()
{
    float goalieXHomeLineMode = -47;
    float goalieXHomeEmg = -49.5;
    float mainGoalieXHomeLM;
	float minGoalieXLM = -50;
	float virtualMinGoalieXLM;

	if (worldModel->getPlayMode() == PM_PENALTY_SETUP ||
		worldModel->getPlayMode() == PM_PENALTY_SETUP_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY)
	{
	}
	else
	{
/////////
//	LATEST TRICK
	if (config["Agent"]["Public"]["LatestTrick"].asBool() &&
		worldModel->getBall().getPos().getX() > -3)
	{
		if (worldModel->getBall().getPos().getX() > -34 &&
			worldModel->getBall().getPos().getX() < -15)
			goalieXHomeLineMode = Basics::reRate(worldModel->getBall().getPos().getX(),
					-25, -15, -47, -38);
		
		if (worldModel->getBall().getPos().getX() > -15 &&
			worldModel->getBall().getPos().getX() < 0)
			goalieXHomeLineMode = Basics::reRate(worldModel->getBall().getPos().getX(),
					-15, 0, -38, -20);

	
		if (worldModel->getBall().getPos().getX() >= 0)
			goalieXHomeLineMode = Basics::reRate(worldModel->getBall().getPos().getX(),
					0, 54, -24, -15);
		if (goalieXHomeLineMode > -32)
			goalieXHomeLineMode = -32;
	}
//	LATEST TRICK 

//////////	
	}	
    Vector upTirToBall, downTirToBall, goalToBall;
    Point ballPoint;
    
    ballPoint.x = worldModel->getBall().getPos().getX() + 
					worldModel->getBall().getVel().getX();
					
    ballPoint.y = worldModel->getBall().getPos().getY() + 
					worldModel->getBall().getVel().getY();

    mainGoalieXHomeLM = goalieXHomeLineMode;
	virtualMinGoalieXLM = -52.5 + float((ballPoint.x - (-52.5)) / 2.8);   
	minGoalieXLM = max(minGoalieXLM, virtualMinGoalieXLM);

	if (worldModel->getBody().getPos().getX() < mainGoalieXHomeLM)
	{
		if (worldModel->getBody().getPos().getX() > minGoalieXLM)
			mainGoalieXHomeLM = worldModel->getBody().getPos().getX();	
	}
	
    upTirToBall.setByPoints(Point(-52.5, 7), ballPoint);
    downTirToBall.setByPoints(Point(-52.5, -7), ballPoint);
    goalToBall.setByPoints(Point(-52.5, 0), ballPoint);
	

    float deltaAngleToUpTir = upTirToBall.getDirection() - 90;
    float deltaAngleToDownTir = downTirToBall.getDirection() - (-90);
    float deltaAngleToGoal; 
   	if (ballPoint.y > 0)
    	deltaAngleToGoal = goalToBall.getDirection() - 90;
	else
    	deltaAngleToGoal = goalToBall.getDirection() - (-90);
		
	
    deltaAngleToUpTir = Degree::normalizeAngle(deltaAngleToUpTir);
    deltaAngleToDownTir = Degree::normalizeAngle(deltaAngleToDownTir);
    deltaAngleToGoal = Degree::normalizeAngle(deltaAngleToGoal);
	
    deltaAngleToUpTir = fabs(deltaAngleToUpTir);
    deltaAngleToDownTir = fabs(deltaAngleToDownTir);
    deltaAngleToGoal = fabs(deltaAngleToGoal);

	LOG << "Goalie LOG 1 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;

    if(worldModel->getBall().getPos().getY() > 0 && 
		deltaAngleToUpTir < 45 && upTirToBall.getMagnitude() < 15)
		{
			mainGoalieXHomeLM = goalieXHomeEmg;
			LOG << "GOALIE Change Function 1" << endl;
			setGoalieParams();
			return;
		}	

	LOG << "Goalie LOG 2 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;
		
    if(worldModel->getBall().getPos().getY() < 0 && 
		deltaAngleToDownTir < 45 && downTirToBall.getMagnitude() < 15)
		{
			mainGoalieXHomeLM = goalieXHomeEmg;
			LOG << "GOALIE Change Function 2" << endl;
			setGoalieParams();
			return;
		}
	LOG << "Goalie LOG 3 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;

    LOG << "deltaAngleToGoal : " << deltaAngleToGoal << endl;
    LOG << "goalToBall.Length : " << goalToBall.getMagnitude() << endl;
	if(deltaAngleToGoal < 45 && goalToBall.getMagnitude() < 30)
		{
			mainGoalieXHomeLM = goalieXHomeEmg;
			LOG << "GOALIE Change Function 3" << endl;
			setGoalieParams();
			return;
		}


    if(goalieXHomeLineMode == goalieXHomeEmg)
	{
		if (fabs(worldModel->getBody().getPos().getX() - 
			mainGoalieXHomeLM) < 1)
		{
	    	mainGoalieXHomeLM = worldModel->getBody().getPos().getX();
	LOG << "Goalie LOG 4 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;
		}
	}
    else if(mainGoalieXHomeLM == goalieXHomeLineMode)
	{
		if(fabs(worldModel->getBody().getPos().getX() - mainGoalieXHomeLM) < 1)
	 	{
	    	mainGoalieXHomeLM = worldModel->getBody().getPos().getX();
	LOG << "Goalie LOG 5 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;
		}
	}
    else if(mainGoalieXHomeLM == goalieXHomeEmg)
    {
		if(fabs(worldModel->getBody().getPos().getX() - mainGoalieXHomeLM) < .5)
	    	mainGoalieXHomeLM = worldModel->getBody().getPos().getX();

	LOG << "Goalie LOG 6 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;
		
		if(fabs(worldModel->getBody().getPos().getX() - mainGoalieXHomeLM) < 1 
			&& worldModel->getBody().getPos().getX() < mainGoalieXHomeLM)
	    	mainGoalieXHomeLM = worldModel->getBody().getPos().getX();

	LOG << "Goalie LOG 7 : " << mainGoalieXHomeLM << " " << goalieXHomeLineMode 
		<< " " << goalieXHomeEmg << endl;
    }	


    Line goalLine, midAngleLine;
    goalLine.setByPoints(Point(mainGoalieXHomeLM, 7), Point(mainGoalieXHomeLM, -7));

    Point tirUpPoint, tirDownPoint;
    Vector tirUpVector, tirDownVector, midAngleVector;
    float deltaAngle, betweenAngle;
    
    tirUpPoint.x = tirDownPoint.x = -52.5;
    tirUpPoint.y = 7;
    tirDownPoint.y = -7;

    tirUpVector.setByPoints(ballPoint, tirUpPoint);
    tirDownVector.setByPoints(ballPoint, tirDownPoint);
    
    deltaAngle = tirUpVector.getDirection() - tirDownVector.getDirection();
    deltaAngle = Degree::normalizeAngle(deltaAngle);

    deltaAngle = fabs(deltaAngle);
    deltaAngle /= 2;
    
    betweenAngle = tirUpVector.getDirection() + deltaAngle;
    betweenAngle = Degree::normalizeAngle(betweenAngle);
    
    midAngleVector.setAsPolar(1, betweenAngle);
    midAngleLine.setBySourceDir(ballPoint, midAngleVector.getDirection());
    
    midAngleLine.getLineIntersect(goalLine, goalieHomePoint);
  
    if(goalieHomePoint.y > 6)
		goalieHomePoint.y = 6;
    
	if(goalieHomePoint.y < -6)
		goalieHomePoint.y = -6;
    
    if(worldModel->getBall().getPos().getY() >= 0)
		mainIdealBetweenAngle = 180;
    
	if(worldModel->getBall().getPos().getY() < 0)
		mainIdealBetweenAngle = 0;
}

Point GoaliePositioning::shootToTirSimulate(Point sourcePoint, 
						Point beforePoint, float idealBodyAngle, 
						float firstDashLength, float maxChangeDistance, 
						float miniChangeDist)
{
    if(fabs(idealBodyAngle) > 90)
    {
		idealBodyAngle += 180;
		idealBodyAngle = Degree::normalizeAngle(idealBodyAngle);
    } 
    
    Body simBody;
    Ball simBall;
    
	simBody = worldModel->getBody();
    simBall = worldModel->getBall();
    
	Point startPoint;
    Vector actionVector;
    float dashPower;
    
    dashPower = 100;
    if(simBody.getStamina() < 1400)
		dashPower = worldModel->getBody().getStaminaIncMax();

	Command *dashCommand = new DashCommand(AT_GOALIE_POSITIONING, dashPower);

    actionVector.setAsPolar(firstDashLength, idealBodyAngle);    

    startPoint.x = sourcePoint.x + actionVector.getX();
    startPoint.y = sourcePoint.y + actionVector.getY();

    simBody.setPos().setAsCartesian(startPoint.x, startPoint.y);
    simBody.setVel().setAsCartesian(0, 0);
    
	simBody.setBodyDir(idealBodyAngle);
    
    Vector shootVector;
    shootVector.setByPoints(simBall.getPos(), sourcePoint); 
    
	simBall.setVel().setAsPolar(2.65, shootVector.getDirection());
    simBall.simulateByDynamics(simBody);
    
//pak nashavad    float distToSourcePoint = hypot(simBall.getPos().getX() - sourcePoint.x, 
//pak nashavad								simBall.getPos().getY() - sourcePoint.y);
    
    while(simBall.getPos().getX() > sourcePoint.x && 
			hypot(simBody.getPos().getX() - sourcePoint.x, 
			simBody.getPos().getY() - sourcePoint.y) < maxChangeDistance) 
    {
		simBall.simulateByDynamics(simBody);
		simBody.simulateByAction(dashCommand);
		simBody.simulateByDynamics();
    }
    
    Point targetPoint;
    targetPoint.x = simBody.getPos().getX();
    targetPoint.y = simBody.getPos().getY();

    float curDistance = hypot(targetPoint.x - 
			worldModel->getBody().getPos().getX(), 
			targetPoint.y - worldModel->getBody().getPos().getY());

    if(curDistance < miniChangeDist)
    {
		targetPoint.x = worldModel->getBody().getPos().getX();
		targetPoint.y = worldModel->getBody().getPos().getY();
    }
	delete dashCommand;
//	if(hypot(targetPoint.x - sourcePoint.x, targetPoint.y - sourcePoint.y) < 
//		hypot(BeforePoint.x - sourcePoint.x, BeforePoint.y - sourcePoint.y))
//		targetPoint = BeforePoint;
    
    return targetPoint;     
}


float GoaliePositioning::getValue(const Library &library)
{
	if (worldModel->getPlayMode() == PM_PENALTY_SETUP ||
		worldModel->getPlayMode() == PM_PENALTY_SETUP_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY)
    	setGoalieParams();
 	else
	    setGoalieParamsLineMode();
    
    float idealBodyAngle = 0;
//pak nashavad	float idealGTPAngle;

    if(worldModel->getBall().getPos().getY() > 0)
    {
		idealBodyAngle = mainIdealBetweenAngle - 90;
		idealBodyAngle = Degree::normalizeAngle(idealBodyAngle);
    }

    if(worldModel->getBall().getPos().getY() <= 0)
    {
		idealBodyAngle = mainIdealBetweenAngle - 90;
		idealBodyAngle = Degree::normalizeAngle(idealBodyAngle);
    }

    float actionDistMin = .08;
    float actionDistMax = 1.5;

    float distMin = 15, distMax = 35;
    float actionDist;
    float playerToBallDist = hypot(worldModel->getBody().getPos().getX() - 
			worldModel->getBall().getPos().getX(), 
			worldModel->getBody().getPos().getY() - 
			worldModel->getBall().getPos().getY());

    if (playerToBallDist < distMin)
		actionDist = actionDistMin;
    else if(playerToBallDist > distMax)
		actionDist = actionDistMax;
    else
    {
		float rangeDistToBall = distMax - distMin;
		float rangeActionDist = actionDistMax - actionDistMin;
		actionDist = actionDistMin + ((playerToBallDist - distMin) / 
						rangeDistToBall) * rangeActionDist;
    }	

    if (worldModel->getBody().getStamina() < 2000)
	{
		actionDist += 1.3;
		actionDist *= 1.2;
	}
	
	if (worldModel->getPlayMode() != PM_PLAY_ON)
	{
		actionDist += 1.3;
		actionDist *= 1.2;
	}
    
	if (worldModel->getPlayMode() == PM_PENALTY_SETUP ||
		worldModel->getPlayMode() == PM_PENALTY_SETUP_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY_OPP ||
		worldModel->getPlayMode() == PM_PENALTY_READY)
		actionDist = .2;	
	
	float distToHome = hypot(worldModel->getBody().getPos().getX() - 
					goalieHomePoint.x, worldModel->getBody().getPos().getY() - 
					goalieHomePoint.y);
	float distToGoal = hypot(worldModel->getBody().getPos().getX() - 
					(-52.5), worldModel->getBody().getPos().getY() - 0);
	if (distToHome >= actionDist)
    {	
		if (distToHome >= 10 || distToGoal > 15)
		{
			if (worldModel->getPlayMode() == PM_FREE_KICK_OPP||
				worldModel->getPlayMode() == PM_OFFSIDE ||
				worldModel->getPlayMode() == PM_FREE_KICK_FAULT ||
				worldModel->getPlayMode() == PM_CATCH_FAULT ||
				worldModel->getPlayMode() == PM_BACK_PASS)
			{
				ForbiddenAreaGotoPoint doGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, worldModel->getBall().
					getPos(), 9.15, worldModel->getBody());
				if (doGotoPoint.getStatus() == 0)
				{
					if (doGotoPoint.getCommand())
						delete doGotoPoint.getCommand();
					goaliePositioningCommand = TurnToPoint(AT_GOALIE_POSITIONING, worldModel->getBall().getPos(),
						worldModel->getBody(), 5).getCommand();
				}
				else
					goaliePositioningCommand = doGotoPoint.getCommand();
			}
			else
				goaliePositioningCommand = PreciseGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, actionDist, 
					worldModel->getBody(), 0).getCommand();
		}
		else
		{
			if (worldModel->getPlayMode() == PM_FREE_KICK_OPP||
				worldModel->getPlayMode() == PM_OFFSIDE ||
				worldModel->getPlayMode() == PM_FREE_KICK_FAULT ||
				worldModel->getPlayMode() == PM_CATCH_FAULT ||
				worldModel->getPlayMode() == PM_BACK_PASS)
			{
				ForbiddenAreaGotoPoint doGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, worldModel->getBall().
					getPos(), 9.15, worldModel->getBody());
				if (doGotoPoint.getStatus() == 0)
				{
					if (doGotoPoint.getCommand())
						delete doGotoPoint.getCommand();
					goaliePositioningCommand = TurnToPoint(AT_GOALIE_POSITIONING, worldModel->getBall().getPos(),
						worldModel->getBody(), 5).getCommand();
				}
				else
					goaliePositioningCommand = doGotoPoint.getCommand();
			}
			else
				goaliePositioningCommand = PreciseGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, actionDist, 
					worldModel->getBody(), 1).getCommand();
		}
    	LOG << "GOALIE PrecideGTP 1 Stamina : " << worldModel->getBody().getStamina() << endl;
	}
	else
    {
		float posIdealAngle, negIdealAngle, deltaPosAngle, 
				deltaNegAngle, maxTurnAngle;

		posIdealAngle = idealBodyAngle;
		negIdealAngle = posIdealAngle - 180;
	   	negIdealAngle = Degree::normalizeAngle(negIdealAngle);

		deltaPosAngle = posIdealAngle - worldModel->getBody().getBodyDir();
		deltaNegAngle = negIdealAngle - worldModel->getBody().getBodyDir();

		deltaPosAngle = Degree::normalizeAngle(deltaPosAngle);
		deltaNegAngle = Degree::normalizeAngle(deltaNegAngle);

		maxTurnAngle = BasicTurn::getTurnAngle(179, worldModel->getBody());

		if (hypot(worldModel->getBody().getPos().getX() - 
			goalieHomePoint.x, worldModel->getBody().getPos().getY() - 
			goalieHomePoint.y) <= actionDist) 
		{
		    if (fabs(deltaPosAngle) - 15 < maxTurnAngle)
				goaliePositioningCommand = TurnAngle(AT_GOALIE_POSITIONING, deltaPosAngle, 
				worldModel->getBody()).getCommand();
	    	else 
	    	{
				if (fabs(deltaNegAngle) - 15 < maxTurnAngle)
		    		goaliePositioningCommand = TurnAngle(AT_GOALIE_POSITIONING, deltaNegAngle, 
					worldModel->getBody()).getCommand();
				else
		    		goaliePositioningCommand = TurnAngle(AT_GOALIE_POSITIONING, deltaNegAngle, 
					worldModel->getBody()).getCommand();
	    	}
		}
		else
		{
	    	if (fabs(deltaPosAngle) - 15 < maxTurnAngle && 
				fabs(deltaPosAngle) > 10)
	    		goaliePositioningCommand = TurnAngle(AT_GOALIE_POSITIONING, deltaNegAngle, 
				worldModel->getBody()).getCommand();
	    	else
 			{
				float distToHome = hypot(worldModel->getBody().getPos().getX() - 
					goalieHomePoint.x, worldModel->getBody().getPos().getY() - 
					goalieHomePoint.y);
				
				float distToGoal = hypot(worldModel->getBody().getPos().getX() - 
					(-52.5), worldModel->getBody().getPos().getY() - 0);
				
				if (distToHome >= 10 || distToGoal > 15)
				{
					if (worldModel->getPlayMode() == PM_FREE_KICK_OPP||
						worldModel->getPlayMode() == PM_OFFSIDE ||
						worldModel->getPlayMode() == PM_FREE_KICK_FAULT ||
						worldModel->getPlayMode() == PM_CATCH_FAULT ||
						worldModel->getPlayMode() == PM_BACK_PASS)
					{
						ForbiddenAreaGotoPoint doGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, worldModel->
							getBall().getPos(), 9.15, worldModel->getBody());
						if (doGotoPoint.getStatus() == 0)
						{
							if (doGotoPoint.getCommand())
								delete doGotoPoint.getCommand();
							goaliePositioningCommand = TurnToPoint(AT_GOALIE_POSITIONING, worldModel->getBall().
									getPos(), worldModel->getBody(), 5).getCommand();
						}
						else
							goaliePositioningCommand = doGotoPoint.getCommand();
					}
					else
						goaliePositioningCommand = PreciseGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, actionDist, 
							worldModel->getBody(), 0).getCommand();
				}
				else
				{
					if (worldModel->getPlayMode() == PM_FREE_KICK_OPP||
						worldModel->getPlayMode() == PM_OFFSIDE ||
						worldModel->getPlayMode() == PM_FREE_KICK_FAULT ||
						worldModel->getPlayMode() == PM_CATCH_FAULT ||
						worldModel->getPlayMode() == PM_BACK_PASS)
					{
						ForbiddenAreaGotoPoint doGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, worldModel->
							getBall().getPos(), 9.15, worldModel->getBody());
						if (doGotoPoint.getStatus() == 0)
						{
							if (doGotoPoint.getCommand())
								delete doGotoPoint.getCommand();
							goaliePositioningCommand = TurnToPoint(AT_GOALIE_POSITIONING, worldModel->getBall().
									getPos(), worldModel->getBody(), 5).getCommand();
						}
						else
							goaliePositioningCommand = doGotoPoint.getCommand();
					}
					else
						goaliePositioningCommand = PreciseGotoPoint(AT_GOALIE_POSITIONING, goalieHomePoint, actionDist, 
							worldModel->getBody(), 1).getCommand();
				}
    			LOG << "GOALIE PrecideGTP 1 Stamina : " << worldModel->getBody().getStamina() << endl;
			}
		}
    }
	return AD_NORMAL_VALUE;
}

void GoaliePositioning::execute(Form &form, const Library &library)
{
	command = goaliePositioningCommand;
}

