/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Meisam Vosoughpour
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>

#include <Degree.h>
#include <Logger.h>
#include <Defines.h>
#include <SRPDribble.h>
#include <InterceptCalculate.h>


//#define SRPDRIBBLE_LOGLEVEL1


using namespace std;

SRPDribble::SRPDribble(const WorldModel *worldModel): AdvancedAction(worldModel)
{
}

SRPDribble::~SRPDribble()
{
}

void SRPDribble::execute(Form &form, const Library &library)
{
	LOG << "SRPDribble::execute" << endl;

	command = srpCommand;
	command->setCreator(AT_SRPDRIBBLE);
}

void SRPDribble::getDangerActionValue()
{
//KOMAKE KHODAM : INHA CHIST?!?    CKickGoToBallPlayer *GetBodyGoToBall, *GetBestOppGoToBall;
    int bodyListNumber, bestOppListNumber;
    
    for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			deltaToFirstTime[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			deltaToFirstTime[i] = -999;
	    	continue;
	    }	

	    bodyListNumber = worldModel->getCurKickCalculate().getSimulate(i).getCheckedListNum(worldModel->getBody());

	    if (bodyListNumber == NOVALUE)
	    {
			deltaToFirstTime[i] = -999;
			sarePaTime[i] = -999;
			continue;
	    }	
	    
	    if (!worldModel->getCurKickCalculate().getSimulate(i).getOppCheckedPlayersNum())
	    {
			deltaToFirstTime[i] = DEFAULT_SHARE_CYCLES;
			sarePaTime[i] = worldModel->getCurKickCalculate().getSimulate(i).
						getCheckedInterPlayer(bodyListNumber).getInterTime();
			oppGtbTime[i] = 999;
	    	continue;
	    }	
	    			
	    bestOppListNumber = MAX_CHECK_PLAYERS + 1;
	    
	    for (unsigned o = 0; o < worldModel->getCurKickCalculate().getSimulate(i).getCheckedPlayersNum(); o++)
	    {
			if(worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(o).getPlayer().getTeamId() == TID_TEAMMATE) 
		 	   continue;
			if (bestOppListNumber == MAX_CHECK_PLAYERS + 1 ||
		    	worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(o).getInterTime() < 
		    	worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bestOppListNumber).getInterTime())
		    	bestOppListNumber = o;	         
	    
	    }
	    
		if (bestOppListNumber == MAX_CHECK_PLAYERS + 1)
		{
			deltaToFirstTime[i] = DEFAULT_SHARE_CYCLES;
		}
		else	
		{
			deltaToFirstTime[i] = (int)(worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bestOppListNumber).getInterTime() -
	                    	     worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bodyListNumber).getInterTime());
		}
		
		if (deltaToFirstTime[i] < 0)
			deltaToFirstTime[i] *= 15;
		else if (deltaToFirstTime == 0)
			deltaToFirstTime[i] -= 4;
		
		sarePaTime[i] = worldModel->getCurKickCalculate().getSimulate(i).
			getCheckedInterPlayer(bodyListNumber).getInterTime();

		oppGtbTime[i] = worldModel->getCurKickCalculate().getSimulate(i).
				getCheckedInterPlayer(bestOppListNumber).getInterTime();
	}
	
}




void SRPDribble::getDangerFieldValue()
{
    int bodyListNumber;
    Point goToBallPoint;
    float dangerX = 50, dangerY = 31.5;
    
    for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    bodyListNumber = worldModel->getCurKickCalculate().getSimulate(i).getCheckedListNum(worldModel->getBody());
	    
		if (bodyListNumber == NOVALUE)
		{
			dangerFieldValue[i] = -1000;
			continue;
		}
				
		goToBallPoint = worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bodyListNumber).getInterPoint();

	    dangerFieldValue[i] = 0;
	    	    
	    if(goToBallPoint.x > dangerX)
			dangerFieldValue[i] -= (goToBallPoint.x - dangerX) / 1;

	    if(fabs(goToBallPoint.y) > dangerY)
			dangerFieldValue[i] -= (fabs(goToBallPoint.y) - dangerY) / 1;
	}
}


void SRPDribble::getSpeedValue()
{
    int bodyListNumber;
    Point goToBallPoint;
    float normalDistance = 8;
    float sarePaPointDistance;
    float normalWeight = 4;                   
    
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    bodyListNumber = worldModel->getCurKickCalculate().getSimulate(i).getCheckedListNum(worldModel->getBody());

		if (bodyListNumber == NOVALUE)
		{
			speedValue[i] = -1000;
			continue;
		}

	    goToBallPoint = worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bodyListNumber).getInterPoint();
	    sarePaPointDistance = hypot(goToBallPoint.x - worldModel->getBall().getPos().getX(), goToBallPoint.y - worldModel->getBall().getPos().getY());
	    
	    if(sarePaPointDistance <= normalDistance)
			speedValue[i] = normalWeight + (sarePaPointDistance - normalDistance);

	    if(sarePaPointDistance > normalDistance)
			speedValue[i] = normalWeight - (sarePaPointDistance - normalDistance) * 6;

	    if (i == 40)
	    {
		LOG << "salam salam sad ta salam : distToSrpPoint : " << sarePaPointDistance << endl;
		LOG << "speedValue[40] : " << speedValue[i] << endl;
	    }
//pak nashavad	    if(speedValue[i] < 0)
//			speedValue[i] = 0;
	    
		Vector gtbVector;
	    float enemySpeedAngle;
	    
		gtbVector.setByPoints(worldModel->getBody().getPos(), goToBallPoint);
	    enemySpeedAngle = gtbVector.getDirection() - worldModel->getBody().getBodyDir();
	    enemySpeedAngle = Degree::normalizeAngle(enemySpeedAngle);
	    enemySpeedAngle = fabs(enemySpeedAngle);
	    if(enemySpeedAngle < 17)
		speedValue[i] += 3.5;
	    if (enemySpeedAngle > 32)	
		speedValue[i] -= 5;
	}
    
}


void SRPDribble::getNextStatusValue()
{
    Point targetPoint;
    targetPoint.x = 51;
    if(worldModel->getBody().getPos().getX() < 36)
		targetPoint.y = worldModel->getBody().getPos().getY();
    else
		targetPoint.y = 0;
	
    int bodyListNumber;
//    Point nextBodyPoint;	
    Vector dribbleTargetVector, goToBallTargetVector;
    float deltaAngleToTarget;
	
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			nextStatusValue[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			nextStatusValue[i] = -999;
	    	continue;
	    }	

	    bodyListNumber = worldModel->getCurKickCalculate().getSimulate(i).getCheckedListNum(worldModel->getBody());

		if (bodyListNumber == NOVALUE)
		{
			nextStatusValue[i] = -999;
			continue;
		}
	    
	    myGoToBallPoint[i] = worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bodyListNumber).getInterPoint();
	    
	    dribbleTargetVector.setByPoints(worldModel->getBody().getPos(), targetPoint);
	    goToBallTargetVector.setByPoints(worldModel->getBody().getPos(), myGoToBallPoint[i]);
	     	
	    deltaAngleToTarget = goToBallTargetVector.getDirection() - dribbleTargetVector.getDirection();
	    deltaAngleToTarget = Degree::normalizeAngle(deltaAngleToTarget);
	    deltaAngleToTarget = fabs(deltaAngleToTarget);
	    
	    nextStatusValue[i] = myGoToBallPoint[i].x - worldModel->getBody().getPos().getX();
	}
}

void SRPDribble::getPathValue()
{
	float firstXPathStep = 42;
	Point targetPoint;
	targetPoint.x = 51;
	if(worldModel->getBody().getPos().getX() < firstXPathStep)
	    targetPoint.y = worldModel->getBody().getPos().getY();
	else
	    targetPoint.y = 0;
    	
    int bodyListNumber;
//    Point nextBodyPoint;	
    Vector dribbleTargetVector, goToBallTargetVector;
    float deltaAngleToTarget;
	
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			pathValue[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			pathValue[i] = -999;
	    	continue;
	    }	

	    bodyListNumber = worldModel->getCurKickCalculate().getSimulate(i).getCheckedListNum(worldModel->getBody());
	
		if (bodyListNumber == NOVALUE)
		{
			pathValue[i] = -999;
			continue;
		}
	    
	    myGoToBallPoint[i] = worldModel->getCurKickCalculate().getSimulate(i).getCheckedInterPlayer(bodyListNumber).getInterPoint();
	    
	    dribbleTargetVector.setByPoints(Point(worldModel->getBody().getPos().getX(), worldModel->getBody().getPos().getY()), targetPoint);
	    goToBallTargetVector.setByPoints(Point(worldModel->getBody().getPos().getX(), worldModel->getBody().getPos().getY()), myGoToBallPoint[i]);
	     	
	    deltaAngleToTarget = goToBallTargetVector.getDirection() - dribbleTargetVector.getDirection();
	    deltaAngleToTarget = Degree::normalizeAngle(deltaAngleToTarget);
	    deltaAngleToTarget = fabs(deltaAngleToTarget);
	    
	    pathValue[i] = (180 - deltaAngleToTarget) / 15.00;  
	
		if (targetPoint.y != 0)
		{
			deltaAngleToTarget = fabs(goToBallTargetVector.getDirection());
	    	pathValue[i] = fmax((180 - deltaAngleToTarget) / 15.00, pathValue[i]);  
		}
	}
    
}

void SRPDribble::getOppBallVectorValue()
{
	float oppDeltaDir;

	Point oppFastestPoint;		
	Vector meToInterPointVector, oppToInterPointVector;
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			oppBallVectorValue[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			oppBallVectorValue[i] = -999;
	    	continue;
	    }	

		if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestOppPlayer())
		{
			oppBallVectorValue[i] = 1;
			continue;
		}
		

		oppFastestPoint = worldModel->getCurKickCalculate().getSimulate(i).
						getFastestOppInterPlayer().getPlayer().getPos();

		meToInterPointVector.setByPoints(worldModel->getBody().getPos(), myGoToBallPoint[i]);
		oppToInterPointVector.setByPoints(oppFastestPoint, myGoToBallPoint[i]);
		
		oppDeltaDir = meToInterPointVector.getDirection() - oppToInterPointVector.getDirection();
		oppDeltaDir = Degree::normalizeAngle(oppDeltaDir);
		oppDeltaDir = fabs(oppDeltaDir);
		
		oppBallVectorValue[i] = pow(Degree::cos(oppDeltaDir), 3);
	}
}

void SRPDribble::getBreakDefenseLineValue()
{
	float oppOffsideLine = worldModel->getOppLocalOffLine();
	if (oppOffsideLine == -0xFFFF)
		oppOffsideLine = worldModel->getOppOffsideLine();

	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			breakDefenseLineValue[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			breakDefenseLineValue[i] = -999;
	    	continue;
	    }	
		if (oppOffsideLine == worldModel->getBall().getPos().getX())
			breakDefenseLineValue[i] = 5.9999;
		else
		{
			breakDefenseLineValue[i] = worldModel->getCurKickCalculate().getSimulate(i).
						getFastestInterPlayer().getInterPoint().x - oppOffsideLine;

			if (breakDefenseLineValue[i] < 0)
				breakDefenseLineValue[i] = 0;
				
			if (breakDefenseLineValue[i] > 8)
				breakDefenseLineValue[i] = 8; 
		}
	}
}

void SRPDribble::getSeeInformationValue()
{
    Vector neckSimVector;

	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
	    {
			seeInformationValue[i] = -999;
			continue;
	    }

	    if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
	    {
			seeInformationValue[i] = -999;
	    	continue;
	    }	
	    neckSimVector.setByPoints(Point(worldModel->getBody().getPos().getX(), worldModel->getBody().getPos().getY()), myGoToBallPoint[i]);

		float seeInfoValue = 0.9323;
		seeInfoValue = worldModel->getSeeDeltaCycle(neckSimVector.getDirection(), 0.5);
		seeInformationValue[i] = 0.333 * (seeInfoValue * seeInfoValue) + seeInfoValue;
	    seeInformationValue[i] *= -1; // It is a big TOF.

	    //////////
	    float deltaAngleToSarePa;
	    deltaAngleToSarePa = worldModel->getBody().getBodyDir() - neckSimVector.getDirection();
	    deltaAngleToSarePa = Degree::normalizeAngle(deltaAngleToSarePa);
	    deltaAngleToSarePa = fabs(deltaAngleToSarePa);

		if(deltaAngleToSarePa > 25)
			seeInformationValue[i] -= (deltaAngleToSarePa / 5) * .1;
	    //////////
	    if(seeInformationValue[i] < -8)
			seeInformationValue[i] = -100;
	}
}

void SRPDribble::getQuantomicStaminaValue()
{
	float subStamina;
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
    	if (!worldModel->getCurKickCalculate().getSimulate(i).isValid())
		{
			quantomicStaminaValue[i] = false;
			continue;
		}
	
    	if (!worldModel->getCurKickCalculate().getSimulate(i).existFastestPlayer())
		{
			quantomicStaminaValue[i] = false;
			continue;
		}
 
	    if (sarePaTime[i] < 0)
		{
			quantomicStaminaValue[i] = false;
			continue;
		}

		subStamina = worldModel->getBody().getStamina() + 
				sarePaTime[i] * worldModel->getBody().getStaminaIncMax() - 
				(sarePaTime[i] - 1) * 100; // 100 = default dash power

#ifdef SRPDRIBBLE_LOGLEVEL1
		LOG << "subStamina of " << i << "th sarePa : " << subStamina << endl;
#endif
		
		if (subStamina < 1400)
			quantomicStaminaValue[i] = false;
		else	
			quantomicStaminaValue[i] = true;
	}
}


float SRPDribble::getValue(const Library &library)
{
	LOG << "SRPDribble::getValue" << endl;

    deltaTimeWeightRate = 3;
    nextStatusWeightRate = 4;
    seeInformationWeightRate = 1.5;
    pathWeightRate = .8;
    dangerFieldWeightRate = 150;
    speedWeightRate = 4;
	breakDefenseLineWeightRate = 4;
	oppBallVectorWeightRate = 0;
    
    getNextStatusValue();
    getDangerActionValue();
    getSeeInformationValue();    
    getPathValue();
    getDangerFieldValue();
    getSpeedValue();
	getBreakDefenseLineValue();
	getOppBallVectorValue();
	getQuantomicStaminaValue();
    
    bestSarePaCounter = 0;
    
	for (unsigned i = 0; i < worldModel->getCurKickCalculate().getSimulatesNum(); i++)
	{
	    sarePaWeight[i] = nextStatusValue[i] * nextStatusWeightRate;
	    sarePaWeight[i] += deltaToFirstTime[i] * deltaTimeWeightRate;      
	    sarePaWeight[i] += seeInformationValue[i] * seeInformationWeightRate;      
	    sarePaWeight[i] += pathValue[i] * pathWeightRate;      
	    sarePaWeight[i] += dangerFieldValue[i] * dangerFieldWeightRate;      
	    sarePaWeight[i] += speedValue[i] * speedWeightRate;      
	    sarePaWeight[i] += breakDefenseLineValue[i] * breakDefenseLineWeightRate;      
	    sarePaWeight[i] += oppBallVectorValue[i] * oppBallVectorWeightRate;      
	    if (!quantomicStaminaValue[i])
			sarePaWeight[i] -= 10001;
		if (sarePaTime[i] > 10)
			sarePaWeight[i] -= pow((sarePaTime[i] - 10), 3);
		if (sarePaTime[i] > 6 && deltaToFirstTime[i] == 0)
			sarePaWeight[i] -= 10001;	  

		if (sarePaTime[i] == 1)
		    sarePaWeight[i] -= 10001;		
		

		if (worldModel->getCurKickCalculate().getSimulate(i).
			isValid() &&
			worldModel->getCurKickCalculate().getSimulate(i).
			existFastestPlayer() &&
			worldModel->getCurKickCalculate().getSimulate(i).
			existFastestOppPlayer())
		{	
#ifdef SRPDRIBBLE_LOGLEVEL1
			
			LOG << "])" << i << "([ FastestOpp : " << worldModel->getCurKickCalculate().getSimulate(i).
			getFastestOppInterPlayer().getPlayer().getPos() << endl;
#endif
		}

		if (worldModel->getCurKickCalculate().getSimulate(i).
			isValid() &&
			worldModel->getCurKickCalculate().getSimulate(i).
			existFastestPlayer() &&
			worldModel->getCurKickCalculate().getSimulate(i).
			existFastestOppPlayer() &&
			worldModel->getCurKickCalculate().getSimulate(i).
			getFastestOppInterPlayer().getPlayer().isGoalie() &&
			deltaToFirstTime[i] / deltaTimeWeightRate <= 1)
		{
#ifdef SRPDRIBBLE_LOGLEVEL1
			LOG << "))" << i << "((opponent goalie is in the SRP path" << endl; 
#endif
			sarePaWeight[i] -= 10000;
		
		}
		if (deltaToFirstTime[i] <= 0)
			sarePaWeight[i] -= 35;

	    if(sarePaWeight[i] > sarePaWeight[bestSarePaCounter])
	    {
			bestSarePaCounter = i;
	    }
#ifdef SRPDRIBBLE_LOGLEVEL1
LOG << "---------------------[)" << i << "(]--------------------" << endl;
//LOG << "Command               : " << 
//	worldModel->getCurKickCalculate().
//	getSimulate(i).getKickCommand().toString() << endl;

LOG << "Point                 : " << 
	myGoToBallPoint[i] << endl;
	
    LOG << "SarePa SeeInformation   : " << 
			seeInformationValue[i] * 
			seeInformationWeightRate << endl;		
    LOG << "SarePa nextStatus       : " << 
			nextStatusValue[i] * 
			nextStatusWeightRate << endl;		
    LOG << "SarePa deltaToFirstTime : " << 
			deltaToFirstTime[i] * 
			deltaTimeWeightRate << endl;		
    LOG << "SarePa sarePaTime       : " << 
			sarePaTime[i] << endl;		
    LOG << "SarePa oppGtbTime       : " << 
			oppGtbTime[i] << endl;		
    LOG << "SarePa path             : " << 
			pathValue[i] * 
			pathWeightRate << endl;		
    LOG << "SarePa dangerField      : " << 
			dangerFieldValue[i] * 
			dangerFieldWeightRate << endl;		
    LOG << "SarePa speed            : " << 
			speedValue[i] * 
			speedWeightRate << endl;		
    LOG << "SarePa breakDefense     : " << 
			breakDefenseLineValue[i] * 
			breakDefenseLineWeightRate << endl;		
    LOG << "SarePa oppBallVector    : " << 
			oppBallVectorValue[i] * 
			oppBallVectorWeightRate << endl;		

    LOG << "SAREPA         TOTAL    : " << 
			sarePaWeight[i] << endl;		
#endif

	}
    
	LOG << "bestSarePaWeight : " << sarePaWeight[bestSarePaCounter] << endl;
	LOG << "bestSarePaCounter : " << bestSarePaCounter << endl;
	LOG << "worldModel->getCurKickCalculate().getSimulatesNum() : "
		<< 	worldModel->getCurKickCalculate().getSimulatesNum() << endl;
	
	if (worldModel->getCurKickCalculate().getSimulate(bestSarePaCounter).isValid())
		srpCommand = new KickCommand(*dynamic_cast<const KickCommand *>
			(&worldModel->getCurKickCalculate().getSimulate(bestSarePaCounter).getKickCommand()));
	else
		srpCommand = NULL;

	if (worldModel->getBall().getPos().getX() > 46)
		sarePaWeight[bestSarePaCounter] -= 10009;

	sarePaWeight[bestSarePaCounter] += oppBallVectorValue[bestSarePaCounter] *
										oppBallVectorWeightRate;

	if (breakDefenseLineValue[bestSarePaCounter] <= 0 && 
		!(worldModel->getBall().getPos().getX() > 25 ||
		((worldModel->getBody().getPostNum() == 6 && 
		worldModel->getBall().getPos().getY() > 15) || 
		(worldModel->getBody().getPostNum() == 7 &&  
		worldModel->getBall().getPos().getY() < -15))))
		sarePaWeight[bestSarePaCounter] = -9090;
	else
		sarePaWeight[bestSarePaCounter] += breakDefenseLineValue[bestSarePaCounter] * 2;

	if (pathValue[bestSarePaCounter] < 10.2)
		sarePaWeight[bestSarePaCounter] -= 35 + (10.2 - pathValue[bestSarePaCounter]) * 20;

	if (oppGtbTime[bestSarePaCounter] == 1 && 
		deltaToFirstTime[bestSarePaCounter] == 0)
	{
		LOG << "the target point of best sarepa counter was in opp kickable" << endl;	
		sarePaWeight[bestSarePaCounter] -= 10000;
	}


//	if (worldModel->getCurKickCalculate().getSimulate(bestSarePaCounter).
//		existFastestPlayer() &&
//		worldModel->getCurKickCalculate().getSimulate(bestSarePaCounter).
//		existFastestOppPlayer() &&
//		worldModel->getCurKickCalculate().getSimulate(bestSarePaCounter).
//		getFastestOppInterPlayer().getPlayer().isGoalie() &&
//		deltaToFirstTime[bestSarePaCounter] <= 1)
//	{
//		LOG << "opp fastest player is goalie" << endl;	
//		sarePaWeight[bestSarePaCounter] -= 100;
//	}

		
	float latestSarePaValue = sarePaWeight[bestSarePaCounter];
	
	
	
	
//	if(worldModel->getBody().getStamina() <= 1800)
//		latestSarePaValue -= 400;

//    if(worldModel->getBody().getStamina() > 1800 && worldModel->getBody().getStamina() <= 2200)
//		latestSarePaValue -= 50;
		
	if (latestSarePaValue < 15)
		latestSarePaValue = -10000;

    return latestSarePaValue;
}

