/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Mohammad Salehe
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cassert>
#include <cmath>
#include <Logger.h>
#include <Types.h>
#include <Command.h>
#include <BasicTurn.h>
#include <BasicKick.h>
#include <BasicDash.h>
#include <Pass.h>
#include <Shoot.h>
#include <NonPlayOn.h>
#include <Offense.h>
#include <Defense.h>
#include <NonPlayOnDecision.h>
#include <Intercept.h>
#include <GoaliePositioning.h>

#include <AdvancedAgent.h>

using namespace std;

#define NON_PLAY_ON_DECISION_LOG1
#define NON_PLAY_ON_DECISION_LOG2

#define FREE_KICK_ADVICER 5

#define GOTOKICK_NUM 0
#define BACKBALL_NUM 1
#define BEFOREKICKOFF_NUM 2
#define DEFENSE_NUM 3
#define OFFENSE_NUM 4 
#define PASS_NUM 5
#define SHOOT_NUM 6
#define INTERCEPT_NUM 7
#define GOALIE_POSITIONING_NUM 8
#define GOALIE_CATCH_OFFENSE_NUM 9

// NOTE : All of player numbers in this file are Uniform Numbers (not Post Numbers)

NonPlayOnDecision::NonPlayOnDecision(const WorldModel *worldModel):
				BodyDecision(worldModel)
{
	addAdvancedAction(new NonPlayOnGotoKick(worldModel), //0
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new NonPlayOnBackBall(worldModel), //1
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new NonPlayOnBeforeKickOff(worldModel), //2
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new Defense(worldModel), //3
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new Offense(worldModel), //4
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new Pass(worldModel), //5
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new Shoot(worldModel), //6
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new Intercept(worldModel), //7
		AdvancedActionWeights(0, 1, 1), 1);
	addAdvancedAction(new GoaliePositioning(worldModel), //8
		AdvancedActionWeights(0, 1, 1), 0);
	addAdvancedAction(new GoalieCatchOffense(worldModel), //9
		AdvancedActionWeights(0, 1, 1), 0);

	dynamic_cast<Pass *>(advancedActions[PASS_NUM])->
			setCurKickCalculate(&worldModel->getCurKickCalculate());
}

NonPlayOnDecision::~NonPlayOnDecision()
{
}

void NonPlayOnDecision::decide(int quickLevel, Form &form,
		const Library &library)
{
	Point positions[11];
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NonPlayOnDecision" << endl;
#endif

	form.headForm.tnMode = TNM_LOOK_NORMALLY;
	form.sayForm.reset();
	if (DefenseOffenseChoice() == 0)
	{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NonPlayOnDecision Returning empty" << endl;
#endif
		command = new EmptyCommand(AT_NONE);
		return;
	}

	resetQuickLevels();
	command = NULL;	
	getCurrentHomePoints(positions, library);

	switch (worldModel->getPlayMode())
	{
	case PM_BEFORE_KICK_OFF:
	case PM_GOAL:
	case PM_GOAL_OPP:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE1" << endl;
#endif
		if (selectBallKicker(positions) == 1)
			dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
				setIsKicker(false);
		else
			dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
				setIsKicker(true);
		advancedActions[BEFOREKICKOFF_NUM]->getValue(library);
		advancedActions[BEFOREKICKOFF_NUM]->execute(form, library);
		command = advancedActions[BEFOREKICKOFF_NUM]->getCommand();
		return;
		
	case PM_KICK_OFF:  
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2" << endl;
#endif
		if (selectBallKicker(positions) == 1)
		{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2 1" << endl;
#endif
			if (selectBallKicker(positions) == 1)
				dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
					setIsKicker(false);
			else
				dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
					setIsKicker(true);
			advancedActions[BEFOREKICKOFF_NUM]->getValue(library);
			advancedActions[BEFOREKICKOFF_NUM]->execute(form, library);
			command = advancedActions[BEFOREKICKOFF_NUM]->getCommand();
			return;
		}
		else
		{
			form.headForm.tnMode = TNM_LOOK_CAREFULLY;
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2 2" << endl;
#endif
			advancedActions[GOTOKICK_NUM]->getValue(library);
			if (dynamic_cast<NonPlayOnGotoKick *>(advancedActions[GOTOKICK_NUM])
					->getStatus() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2 3" << endl;
#endif
				advancedActions[GOTOKICK_NUM]->execute(form, library);
				command = advancedActions[GOTOKICK_NUM]->getCommand();
				return;
			}
			else if (backBallChoice() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2 4" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				quickLevels[PASS_NUM] = 1;
				quickLevels[SHOOT_NUM] = 1;
				form.sayForm.reset();

				BodyDecision::decide(quickLevel, form, library);
				return;
			}
			else
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE2 5" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				advancedActions[BACKBALL_NUM]->getValue(library);
				advancedActions[BACKBALL_NUM]->execute(form, library);
				command = advancedActions[BACKBALL_NUM]->getCommand();
				return;
			}
		}
		break;

	case PM_KICK_OFF_OPP:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE3" << endl;
#endif
		if (selectBallKicker(positions) == 1)
			dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
				setIsKicker(false);
		else
			dynamic_cast<NonPlayOnBeforeKickOff *>(advancedActions[BEFOREKICKOFF_NUM])->
				setIsKicker(true);
		advancedActions[BEFOREKICKOFF_NUM]->getValue(library);
		advancedActions[BEFOREKICKOFF_NUM]->execute(form, library);
		command = advancedActions[BEFOREKICKOFF_NUM]->getCommand();
		return;
		break;

	case PM_FREE_KICK:
	case PM_KICK_IN:
	case PM_CORNER_KICK:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4" << endl;
#endif
		selectBallKickerSay(positions, form);
		if (selectBallKicker(positions) == 1)
		{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 1" << endl;
#endif
			if (worldModel->getBody().isGoalie())
			{
				advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
				advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
				command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
				return;
			}
			else
			{	
				if (!worldModel->getBody().isGoalie() &&
					worldModel->getPlayMode() == PM_FREE_KICK &&
					worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
				{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 1 2" << endl;
#endif
					advancedActions[GOALIE_CATCH_OFFENSE_NUM]->getValue(library);
					advancedActions[GOALIE_CATCH_OFFENSE_NUM]->
						execute(form, library);
					command = advancedActions[GOALIE_CATCH_OFFENSE_NUM]->
						getCommand();
					return;
				}
				advancedActions[OFFENSE_NUM]->getValue(library);
				advancedActions[OFFENSE_NUM]->execute(form, library);
				command = advancedActions[OFFENSE_NUM]->getCommand();
				return;
			}
		}
		else
		{
			form.headForm.tnMode = TNM_LOOK_CAREFULLY;
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 2" << endl;
#endif
			if (worldModel->getBody().isGoalie() &&
				worldModel->getPlayMode() == PM_FREE_KICK &&
				worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 3 1" << endl;
#endif
				if (backBallChoice() == 1)
				{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 3 1 1" << endl;
#endif
					advancedActions[GOALIE_CATCH_OFFENSE_NUM]->getValue(library);
					advancedActions[GOALIE_CATCH_OFFENSE_NUM]->execute(form, library);
					if (dynamic_cast<GoalieCatchOffense *>
							(advancedActions[GOALIE_CATCH_OFFENSE_NUM])->
							getStatus() == 1)
					{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 3.2" << endl;
#endif
						command = advancedActions[GOALIE_CATCH_OFFENSE_NUM]->
							getCommand();
						return;
					}
					else
					{
						if (worldModel->getPlayMode() == PM_FREE_KICK &&
							worldModel->getVirtualPlayMode() == 
							VPM_GOALIE_CATCH_BALL)
						{
							Point virtualPassPoint(-33, 28);
							
							LOG << "NPP FKick to " << virtualPassPoint << endl;
							command = FastKickToPoint(AT_PASS, virtualPassPoint,
							worldModel->getBall(), worldModel->getBody()).
								getCommand();
							LOG << "NPO cmd:" << command->toString() << endl;

							form.sayForm.interceptNum = 6;
							form.sayForm.beforePassSay = true;
							form.sayForm.freeKickSay = false;
							return;
						}
						delete advancedActions[GOALIE_CATCH_OFFENSE_NUM]->getCommand();
						quickLevels[PASS_NUM] = 1;
						form.sayForm.reset();

						BodyDecision::decide(quickLevel, form, library);
						return;
					}
				}
				else
				{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 3 1 2" << endl;
#endif
					advancedActions[BACKBALL_NUM]->getValue(library);
					advancedActions[BACKBALL_NUM]->execute(form, library);
					command = advancedActions[BACKBALL_NUM]->getCommand();
					return;
				}
				break;
			}
			advancedActions[GOTOKICK_NUM]->getValue(library);
			if (dynamic_cast<NonPlayOnGotoKick *>(advancedActions[GOTOKICK_NUM])
					->getStatus() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 3" << endl;
#endif
				advancedActions[GOTOKICK_NUM]->execute(form, library);
				command = advancedActions[GOTOKICK_NUM]->getCommand();
				return;
			}
			else if (backBallChoice() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 4" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				quickLevels[PASS_NUM] = 1;
				quickLevels[SHOOT_NUM] = 1;
				if (worldModel->getPlayMode() == PM_FREE_KICK &&
					worldModel->getVirtualPlayMode() == VPM_INDIRECT_FREE_KICK)
				{
					quickLevels[SHOOT_NUM] = -1;
				}
				form.sayForm.reset();

				BodyDecision::decide(quickLevel, form, library);
				return;
			}
			else
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE4 5" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				advancedActions[BACKBALL_NUM]->getValue(library);
				advancedActions[BACKBALL_NUM]->execute(form, library);
				command = advancedActions[BACKBALL_NUM]->getCommand();
				return;
			}
		}
		break;

	case PM_FREE_KICK_OPP:
	case PM_KICK_IN_OPP:
	case PM_CORNER_KICK_OPP:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE5" << endl;
#endif
		if (worldModel->getBody().isGoalie())
		{
			advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
			advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
			command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
			return;
		}
		else
		{	
			advancedActions[DEFENSE_NUM]->getValue(library);
			advancedActions[DEFENSE_NUM]->execute(form, library);
			command = advancedActions[DEFENSE_NUM]->getCommand();
			return;
		}
		break;

	case PM_OFFSIDE:
	case PM_FREE_KICK_FAULT:
	case PM_CATCH_FAULT:
	case PM_BACK_PASS:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE6" << endl;
#endif
		if (worldModel->getBody().isGoalie())
		{
			advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
			advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
			command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
			return;
		}
		else
		{	
			advancedActions[DEFENSE_NUM]->getValue(library);
			advancedActions[DEFENSE_NUM]->execute(form, library);
			command = advancedActions[DEFENSE_NUM]->getCommand();
			return;
		}
		break;

	case PM_OFFSIDE_OPP:
	case PM_FREE_KICK_FAULT_OPP:
	case PM_CATCH_FAULT_OPP:
	case PM_BACK_PASS_OPP:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7" << endl;
#endif
		selectBallKickerSay(positions, form);
		if (selectBallKicker(positions) == 1)
		{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7 1" << endl;
#endif
			if (worldModel->getBody().isGoalie())
			{
				advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
				advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
				command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
				return;
			}
			else
			{	
				advancedActions[OFFENSE_NUM]->getValue(library);
				advancedActions[OFFENSE_NUM]->execute(form, library);
				command = advancedActions[OFFENSE_NUM]->getCommand(); 
				return;
			}
		}
		else
		{
			form.headForm.tnMode = TNM_LOOK_CAREFULLY;
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7 2" << endl;
#endif
			advancedActions[GOTOKICK_NUM]->getValue(library);
			if (dynamic_cast<NonPlayOnGotoKick *>(advancedActions[GOTOKICK_NUM])
					->getStatus() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7 3" << endl;
#endif
				advancedActions[GOTOKICK_NUM]->execute(form, library);
				command = advancedActions[GOTOKICK_NUM]->getCommand();
				return;
			}
			else if (backBallChoice() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7 4" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				command = new EmptyCommand(AT_NONE);
				return;
			}
			else
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE7 5" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				advancedActions[BACKBALL_NUM]->getValue(library);
				advancedActions[BACKBALL_NUM]->execute(form, library);
				return;
			}
		}
		break;

		
	case PM_GOAL_KICK:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8" << endl;
#endif
		if (worldModel->getBody().isGoalie())
		{
			form.headForm.tnMode = TNM_LOOK_CAREFULLY;
			if (worldModel->getBall().getPos().getY() > 0)
				form.sayForm.attentionNum = 7;
			else
				form.sayForm.attentionNum = 8;
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8 1" << endl;
#endif
			if (worldModel->getVirtualPlayMode() == VPM_GOAL_KICK_TAKEN)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8 1 0,5" << endl;
#endif
				advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
				advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
				command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
				return;
			}
			
			advancedActions[GOTOKICK_NUM]->getValue(library);
			if (dynamic_cast<NonPlayOnGotoKick *>(advancedActions[GOTOKICK_NUM])
					->getStatus() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8 1 1" << endl;
#endif
				advancedActions[GOTOKICK_NUM]->execute(form, library);
				command = advancedActions[GOTOKICK_NUM]->getCommand();
				return;
			}
			else if (backBallChoice() == 1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8 1 2" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				quickLevels[PASS_NUM] = 1;
				quickLevels[SHOOT_NUM] = 1;
				form.sayForm.reset();

				BodyDecision::decide(quickLevel, form, library);
				return;
			}
			else
			{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE8 1 3" << endl;
#endif
				// No need to the following line
				if (advancedActions[GOTOKICK_NUM]->getCommand())
					delete advancedActions[GOTOKICK_NUM]->getCommand();
				advancedActions[BACKBALL_NUM]->getValue(library);
				advancedActions[BACKBALL_NUM]->execute(form, library);
				command = advancedActions[BACKBALL_NUM]->getCommand();
				return;
			}
		}
		else
		{
			if (worldModel->getBall().getPos().getY() > 0)
			{
				if (worldModel->getBody().getUniNum() == 7)
				{
					if (worldModel->getVirtualPlayMode() == VPM_GOAL_KICK_TAKEN)
						;
//						form.sayForm.reset();
					else
						form.sayForm.attentionNum = 1;
				}
				else
				{
					if (worldModel->getVirtualPlayMode() == VPM_GOAL_KICK_TAKEN)
						;
//						form.sayForm.reset();
					else
						form.sayForm.attentionNum = 1;
				}
			}
			else
			{
				if (worldModel->getBody().getUniNum() == 8)
				{
					if (worldModel->getVirtualPlayMode() == VPM_GOAL_KICK_TAKEN)
						;
//						form.sayForm.reset();
					else
						form.sayForm.attentionNum = 1;
				}
				else
				{
					if (worldModel->getVirtualPlayMode() == VPM_GOAL_KICK_TAKEN)
						;
//						form.sayForm.reset();
					else
						form.sayForm.attentionNum = 1;
				}
			}

			advancedActions[GOALIE_CATCH_OFFENSE_NUM]->getValue(library);
			advancedActions[GOALIE_CATCH_OFFENSE_NUM]->
				execute(form, library);
			command = advancedActions[GOALIE_CATCH_OFFENSE_NUM]->
				getCommand();
			return;
		}
		break;
		
	case PM_GOAL_KICK_OPP:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE9" << endl;
#endif
		if (worldModel->getBall().getVel().getMagnitude() < 0.1)
		{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE9 1" << endl;
#endif
			if (worldModel->getBody().isGoalie())
			{
				advancedActions[GOALIE_POSITIONING_NUM]->getValue(library);
				advancedActions[GOALIE_POSITIONING_NUM]->execute(form, library);
				command = advancedActions[GOALIE_POSITIONING_NUM]->getCommand();
				return;
			}
			else
			{
				advancedActions[DEFENSE_NUM]->getValue(library);
				advancedActions[DEFENSE_NUM]->execute(form, library);
				command = advancedActions[DEFENSE_NUM]->getCommand();
				return;
			}
		}
		else
		{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE9 2" << endl;
#endif
			if (worldModel->getCurInterCalculate().existFastestTmmPlayer() &&
				worldModel->getCurInterCalculate().getFastestTmmPlayer().isBody())
			{
				Point interceptStandPoint;

				if (abs(worldModel->getBall().getVel().getX()) <= EPS)
				{
					if (worldModel->getBall().getVel().getY() >= 0)
						interceptStandPoint = Point(worldModel->getBall().getPos().getX(), 20);
					else
						interceptStandPoint = Point(worldModel->getBall().getPos().getX(), -20);
				}
				else if (abs(worldModel->getBall().getVel().getY()) <= EPS)
				{
					interceptStandPoint = Point(35.5, worldModel->getBall().getPos().getY());
				}
				else
				{
					float m, b, t;
					m = worldModel->getBall().getVel().getY() /
						worldModel->getBall().getVel().getX();
					b = worldModel->getBall().getPos().getY() -
						worldModel->getBall().getPos().getX() * m;

					t = m * 35.5 + b;
					if (abs(t) <= 20.5)
					{
						interceptStandPoint = Point(35.5, t);
					}
					else
					{
						if (worldModel->getBall().getVel().getY() >= 0)
						{
							t = (20.5 - b) / m;
							if (t > 52)
								t = 52;
							interceptStandPoint = Point(t, 20.5);
						}
						else
						{
							t = (-20.5 - b) / m;
							if (t > 52)
								t = 52;
							interceptStandPoint = Point(t, -20.5);
						}
					}
					
				}
				
				command = ForbiddenAreaGotoPoint(AT_INTERCEPT, interceptStandPoint,
				   	36, -20, 20, worldModel->getBody()).getCommand();
				return;
/*
//TODO:			advancedActions[INTERCEPT_NUM].setForbiddenArea(OPPPENALTYAREA);
				advancedActions[INTERCEPT_NUM]->getValue();
				advancedActions[INTERCEPT_NUM]->execute(sayFlags);
				command = advancedActions[INTERCEPT_NUM]->getCommand();
				return advancedActions[INTERCEPT_NUM]->getType();
*/
			}
		}
		break;
	case PM_TIME_UP:
	case PM_TIME_UP_WITHOUT_A_TEAM:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE-1 TIMEOVER(CHECK IT)" << endl;
#endif
		command = new EmptyCommand(AT_NONE);
		return;
		break;	
	default:
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "NONPLAYON: MODE0 default(CHECK IT)" << endl;
#endif
			command = new EmptyCommand(AT_NONE);
			return;
	}
	
	return;
}

int NonPlayOnDecision::selectBallKicker(Point positions[11])
{
	float nearestDistance;
	unsigned i;

	unsigned actualPlayModeStayCycles,actualPlayModeRemainCycles;

	Point myHome;
	Vector meBall;
	Vector homeBall;

#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "selectKicker : start" << endl;
#endif
	if (worldModel->getBody().isGoalie() &&
		worldModel->getPlayMode() == PM_FREE_KICK &&
		worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
		return 2;

 	if (!worldModel->getBody().isGoalie() &&
		worldModel->getPlayMode() == PM_FREE_KICK &&
		worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
		return 1;

	if (worldModel->getBody().isGoalie() &&
		worldModel->getPlayMode() == PM_GOAL_KICK &&
		worldModel->getVirtualPlayMode() != VPM_GOAL_KICK_TAKEN)
		return 2;

 	if (!worldModel->getBody().isGoalie() &&
		worldModel->getPlayMode() == PM_GOAL_KICK &&
		worldModel->getVirtualPlayMode() != VPM_GOAL_KICK_TAKEN)
		return 1;

	if (worldModel->getBody().isGoalie())
		return 1;

	if (worldModel->getPlayMode() == PM_CORNER_KICK)
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : corner kick" << endl;
#endif
		if (worldModel->getBall().getPos().getY() > 0)
		{
			if (worldModel->getBody().getUniNum() == 10)
				return 2;
			else
				return 1;
		}
		else
		{
			if (worldModel->getBody().getUniNum() == 11)
				return 2;
			else
				return 1;
		}
	}

	if (worldModel->getPlayMode() == PM_BEFORE_KICK_OFF ||
		worldModel->getPlayMode() == PM_GOAL ||
		worldModel->getPlayMode() == PM_GOAL_OPP ||
		worldModel->getPlayMode() == PM_KICK_OFF ||
		worldModel->getPlayMode() == PM_KICK_OFF_OPP)
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : simple" << endl;
#endif
		nearestDistance = 10000;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
		{
			if (i + 1 == GOALIE_UNIFORM_NUM) continue; // Goalie
			homeBall.setByPoints(positions[i], worldModel->getBall().getPos());
	
			if (homeBall.getMagnitude() < nearestDistance)
				nearestDistance = homeBall.getMagnitude();
		}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : nearestdist" << nearestDistance << endl;
#endif
		myHome = positions[worldModel->getBody().getUniNum() - 1];
		homeBall.setByPoints(myHome, worldModel->getBall().getPos());

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : mydist" << homeBall.getMagnitude() << endl;
#endif
		if (homeBall.getMagnitude() < nearestDistance + EPS)
			return 2;
		else
			return 1;
	}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : normal" << endl;
#endif

	actualPlayModeRemainCycles = worldModel->getPlayModeRemainCycle();
   	actualPlayModeStayCycles = (unsigned)worldModel->getServerParam()["drop_ball_time"].asInt();
	if (worldModel->getPlayMode() == PM_OFFSIDE_OPP ||
		worldModel->getPlayMode() == PM_FREE_KICK_FAULT_OPP ||
		worldModel->getPlayMode() == PM_CATCH_FAULT_OPP)
	{
		return 1;
	}
	
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : actualPlayModeRemainCycles" << actualPlayModeRemainCycles << endl;
LOG << "selectKicker : actualPlayModeStayCycles" << actualPlayModeStayCycles << endl;
#endif

	if (actualPlayModeRemainCycles > actualPlayModeStayCycles - 6) 
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : wait cycles (return 1)" << endl;
#endif
		return 1;
	}

	nearestDistance = 10000;
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		if (i + 1 == GOALIE_UNIFORM_NUM) continue; // Goalie
		if (i + 1 == worldModel->getBody().getUniNum()) continue; // Body
		homeBall.setByPoints(positions[i], worldModel->getBall().getPos());

		if (homeBall.getMagnitude() < nearestDistance)
			nearestDistance = homeBall.getMagnitude();
	}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : nearestdist(w/o body)" << nearestDistance << endl;
#endif
	myHome = positions[worldModel->getBody().getUniNum() - 1];
	homeBall.setByPoints(myHome, worldModel->getBall().getPos());

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : mydist" << homeBall.getMagnitude() << endl;
#endif
	if (homeBall.getMagnitude() > nearestDistance + 20)
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : normal very far(return 1)" << endl;
#endif
		return 1;
	}


	if (worldModel->getFreeKickAdvicerSayCycle() <= worldModel->getCurTime() -
			(actualPlayModeStayCycles - actualPlayModeRemainCycles))
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : late advice(return 1)" << endl;
#endif
		return 1;
	}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : listening to advice" << worldModel->getFreeKickAdvicerSayNumber() << endl;
#endif

	if (worldModel->getFreeKickAdvicerSayNumber() == worldModel->getBody().getUniNum())
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : listening to advice i go" << endl;
#endif
		return 2;
	}
	else
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKicker : listening to advice i don't go" << endl;
#endif
		return 1;
	}

	return 1;
}

void NonPlayOnDecision::selectBallKickerSay(Point positions[11], Form &form)
{
	Point myHome;
	Vector homeBall;
	
	float nearestDistance;
	unsigned nearestPlayer;
	unsigned i;

	unsigned actualPlayModeStayCycles,actualPlayModeRemainCycles;

	actualPlayModeRemainCycles = worldModel->getPlayModeRemainCycle();
   	actualPlayModeStayCycles = (unsigned)worldModel->getServerParam()["drop_ball_time"].asInt();
	if (worldModel->getPlayMode() == PM_OFFSIDE_OPP ||
		worldModel->getPlayMode() == PM_FREE_KICK_FAULT_OPP ||
		worldModel->getPlayMode() == PM_CATCH_FAULT_OPP)
	{
		return;
	}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : actualPlayModeRemainCycles" << actualPlayModeRemainCycles << endl;
LOG << "selectKickerSay : actualPlayModeStayCycles" << actualPlayModeStayCycles << endl;
#endif

	form.sayForm.freeKickSay = false;
	if (actualPlayModeRemainCycles >= actualPlayModeStayCycles - 1)
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : reset savedAdvicePlayer" << endl;
#endif
		savedAdvicePlayer = -1;
	}

	if (actualPlayModeRemainCycles < actualPlayModeStayCycles - 1 &&
		worldModel->getBall().getSeeTime() > worldModel->getCurTime() -
		(actualPlayModeStayCycles - actualPlayModeRemainCycles))
	{
		if (worldModel->getBody().getUniNum() == FREE_KICK_ADVICER)
		{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : advice" << endl;
LOG << "selectKickerSay : current advice=" << savedAdvicePlayer << endl;
#endif
			if (savedAdvicePlayer == -1)
			{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : find advice" << endl;
#endif
				if (!worldModel->getBody().isGoalie() &&
					worldModel->getPlayMode() == PM_FREE_KICK &&
					worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
				{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : find advice" << endl;
#endif

					savedAdvicePlayer = 1;
				}
				else if (worldModel->getPlayMode() == PM_CORNER_KICK)
				{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : corner kick" << endl;
#endif
					if (worldModel->getBall().getPos().getY() > 0)
						savedAdvicePlayer = 10;
					else
						savedAdvicePlayer = 11;
				}
				else if (worldModel->getPlayMode() == PM_KICK_IN &&
						worldModel->getBall().getPos().getX() >= 35)
				{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : near corner kick" << endl;
#endif
					if (worldModel->getBall().getPos().getY() > 0)
						savedAdvicePlayer = 10;
					else
						savedAdvicePlayer = 11;
				}
				else
				{
					nearestDistance = 100000;
					nearestPlayer = worldModel->getBody().getUniNum();
					for (i = 0; i < FULL_PLAYERS_NUM; i++)
					{
						if (i + 1 == GOALIE_UNIFORM_NUM) continue; // Goalie
						homeBall.setByPoints(positions[i], worldModel->getBall().getPos());

						if (homeBall.getMagnitude() < nearestDistance)
						{
							nearestPlayer = i + 1;
							nearestDistance = homeBall.getMagnitude();
						}
					}
					savedAdvicePlayer = nearestPlayer;
				}
			}
			form.sayForm.freeKickSay = true;
			form.sayForm.freeKickAdviceNumber = savedAdvicePlayer;
			form.sayForm.attentionNum = savedAdvicePlayer;

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : finded advice=" << savedAdvicePlayer << endl;
#endif
		}
		else
		{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : listener" << endl;
#endif
			if (worldModel->getFreeKickAdvicerSayCycle() > worldModel->getCurTime() -
				(actualPlayModeStayCycles - actualPlayModeRemainCycles))
			{
				form.sayForm.freeKickSay = true;
				form.sayForm.freeKickAdviceNumber = worldModel->getFreeKickAdvicerSayNumber();
				form.sayForm.attentionNum = worldModel->getFreeKickAdvicerSayNumber();
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : listener listened=" << form.sayForm.freeKickAdviceNumber << endl;
#endif
			}
			else
			{
				form.sayForm.attentionNum = FREE_KICK_ADVICER;
			}
		}
	}
	else
	{
#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "selectKickerSay : 1 cycle after changePlayMode or late Ball" << endl;
#endif
	}

	// Previous Method
}

int NonPlayOnDecision::backBallChoice(int cyclesToWait)
{
	unsigned waitMinCycles;
	int i;
	waitMinCycles = 32;

	unsigned actualPlayModeStayCycles,actualPlayModeRemainCycles;

	actualPlayModeRemainCycles = worldModel->getPlayModeRemainCycle();
   	actualPlayModeStayCycles = (unsigned)worldModel->getServerParam()["drop_ball_time"].asInt();
	if (worldModel->getPlayMode() == PM_OFFSIDE_OPP ||
		worldModel->getPlayMode() == PM_FREE_KICK_FAULT_OPP ||
		worldModel->getPlayMode() == PM_CATCH_FAULT_OPP)
	{
		actualPlayModeRemainCycles += worldModel->getServerParam()["drop_ball_time"].asInt();
		actualPlayModeStayCycles += 30;
	}

#ifdef NON_PLAY_ON_DECISION_LOG2
LOG << "backBallChoice : actualPlayModeRemainCycles" << actualPlayModeRemainCycles << endl;
LOG << "backBallChoice : actualPlayModeStayCycles" << actualPlayModeStayCycles << endl;
#endif

	if (worldModel->getPlayMode() == PM_FREE_KICK &&
		worldModel->getVirtualPlayMode() == 
		VPM_GOALIE_CATCH_BALL)
		waitMinCycles = 100;

	if (actualPlayModeRemainCycles < 3)
		return 1;
	if (actualPlayModeRemainCycles > actualPlayModeStayCycles - waitMinCycles)
		return 2;
	if (cyclesToWait > 0)
		return 2;
	if (worldModel->getBody().getStamina() <
			worldModel->getBody().getStaminaMax() - 1)
		return 2;

	// Check Other Players Stamina
	for (i = 0; i < 11; i++)
		if (worldModel->getFullPlayer(TID_TEAMMATE, i).isValid() &&
			worldModel->getFullPlayer(TID_TEAMMATE, i).getStamina() < 
			worldModel->getFullPlayer(TID_TEAMMATE, i).getStaminaMax() - 1)
		{
			LOG << "backBallChoice : COPS player " << i + 1 << endl;
			return 2; // FIXME : turned off because of ...
		}
	
	return 1;
}

void NonPlayOnDecision::getCurrentHomePoints(Point positions[11],
		const Library &library)
{
	int i;
	if (worldModel->getPlayMode() == PM_BEFORE_KICK_OFF ||
		worldModel->getPlayMode() == PM_KICK_OFF ||
		worldModel->getPlayMode() == PM_KICK_OFF_OPP)
	{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "getMovePoint" << endl;
#endif
		for (i = 1; i <= FULL_PLAYERS_NUM; i++)
			positions[i - 1] = (dynamic_cast<NonPlayOnBeforeKickOff *>
				(advancedActions[BEFOREKICKOFF_NUM]))->getMovePoint(i);
	}
	else if (worldModel->getPlayMode() == PM_FREE_KICK &&
			worldModel->getVirtualPlayMode() == VPM_GOALIE_CATCH_BALL)
	{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "GoalieCatchOffense" << endl;
#endif
		advancedActions[GOALIE_CATCH_OFFENSE_NUM]->getValue(library);
		for (i = 1; i <= FULL_PLAYERS_NUM; i++)
			positions[i - 1] = (dynamic_cast<GoalieCatchOffense *>
				(advancedActions[GOALIE_CATCH_OFFENSE_NUM]))->getHomePoint(i);
	}
	else if (DefenseOffenseChoice() == 2)
	{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "Offense.getHomePoint" << endl;
#endif
		advancedActions[OFFENSE_NUM]->getValue(library);
		for (i = 1; i <= FULL_PLAYERS_NUM; i++)
			positions[i - 1] = (dynamic_cast<Offense *>
				(advancedActions[OFFENSE_NUM]))->getHomePoint(i, library);
	}
	else
	{
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "Defense.getHomePoint" << endl;
#endif
		advancedActions[DEFENSE_NUM]->getValue(library);
		for (i = 1; i <= FULL_PLAYERS_NUM; i++)
			positions[i - 1] = (dynamic_cast<Defense *>
				(advancedActions[DEFENSE_NUM]))->getHomePoint(i);
	}
#ifdef NON_PLAY_ON_DECISION_LOG1
LOG << "All home points found." << endl;
#endif
#ifdef NON_PLAY_ON_DECISION_LOG2
for (i = 1; i <= FULL_PLAYERS_NUM; i++)
	LOG << "homepoint[" << i << "] = " << positions[i - 1] << endl;

#endif
}

int NonPlayOnDecision::DefenseOffenseChoice()
{
	switch (worldModel->getPlayMode())
	{
	case PM_KICK_OFF:
	case PM_FREE_KICK:
	case PM_KICK_IN:
	case PM_OFFSIDE_OPP:
	case PM_GOAL_OPP:
	case PM_GOAL_KICK:
	case PM_CORNER_KICK:
	case PM_FREE_KICK_FAULT_OPP:
	case PM_CATCH_FAULT_OPP:
	case PM_BACK_PASS_OPP:
		return 2; // Offense
		break;
		
	case PM_KICK_OFF_OPP:
	case PM_FREE_KICK_OPP:
	case PM_KICK_IN_OPP:
	case PM_OFFSIDE:
	case PM_GOAL:
	case PM_GOAL_KICK_OPP:
	case PM_CORNER_KICK_OPP:
	case PM_FREE_KICK_FAULT:
	case PM_CATCH_FAULT:
	case PM_BACK_PASS:
		return 1; // Defense
		break;
	case PM_TIME_UP_WITHOUT_A_TEAM:
	case PM_TIME_UP:
		return 0; // NoWork 
		break;
	default:
		return 1; // Defense
		break;
	}
}

