/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Darioush Jalali, Mohammad Salehe
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <AdvancedAgent.h>
#include <Logger.h>
#include <Basics.h>
#include <InterceptCalculate.h>
#include <cmath>
#include <GoalStraightPosPlan.h>
#include <Line.h>
#include <map>
#include <algorithm>
#include <Degree.h>
#include <BasicDash.h>
#include <BasicTurn.h>
#include <cassert>

using namespace std;
using namespace Basics;
using namespace Degree;


#define LOG_INDIVIDUAL_POINT 0
#define MAX_PATH_POINTS 20
#define PATH_ABS_ANGLE 70

GoalStraightPosPlan::GoalStraightPosPlan(const WorldModel *wm, Form &form,
                                         const Library &library): Plan("GoalPos", wm, form, library),
																																	virtualPass(wm)
{
		config.add("GoalStraightPos");
		config["GoalStraightPos"].setByFile("./Configs/GoalStraightPos.conf");
		LOG << "Loading goal positioning factors from conf..." << endl;
		setPassingPaths();
}


void
GoalStraightPosPlan::setPassingPaths() 
{
		// passing paths..
		LOG << "setting passing paths..." << endl;
		passingPaths[make_pair(10,2)] = config["GoalStraightPos"]["PassingPaths"]["Values"][0].asFloat();
		passingPaths[make_pair(2,10)] = config["GoalStraightPos"]["PassingPaths"]["Values"][1].asFloat();
		passingPaths[make_pair(2,3)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][2].asFloat();
		passingPaths[make_pair(3,2)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][3].asFloat();
		passingPaths[make_pair(3,4)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][4].asFloat();
		passingPaths[make_pair(4,3)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][5].asFloat();
		passingPaths[make_pair(2,4)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][6].asFloat();
		passingPaths[make_pair(4,2)]  = config["GoalStraightPos"]["PassingPaths"]["Values"][7].asFloat();
		passingPaths[make_pair(4,11)] = config["GoalStraightPos"]["PassingPaths"]["Values"][8].asFloat();
		passingPaths[make_pair(11,4)] = config["GoalStraightPos"]["PassingPaths"]["Values"][9].asFloat();
		passingPaths[make_pair(10,3)] = config["GoalStraightPos"]["PassingPaths"]["Values"][10].asFloat();				
		passingPaths[make_pair(3,10)] = config["GoalStraightPos"]["PassingPaths"]["Values"][11].asFloat();
		passingPaths[make_pair(11,3)] = config["GoalStraightPos"]["PassingPaths"]["Values"][12].asFloat();				
		passingPaths[make_pair(3,11)] = config["GoalStraightPos"]["PassingPaths"]["Values"][13].asFloat();
		passingPaths[make_pair(10,11)]= config["GoalStraightPos"]["PassingPaths"]["Values"][14].asFloat();				
		passingPaths[make_pair(11,10)]= config["GoalStraightPos"]["PassingPaths"]["Values"][15].asFloat();				
}



GoalStraightPosPlan::~GoalStraightPosPlan()
{
}


void
GoalStraightPosPlan::decide()
{
		LOG << "GoalStraightPosPlan::decide" <<endl;
		if ((isInPlayers(wm->getBody().getUniNum(), "A") &&
				 wm->getHearedPlanStatus() == 1) ||
        (isInPlayers(wm->getBody().getUniNum(), "B") &&
         wm->getHearedPlanStatus() == 2))
        virtualPass.manager(form, library);
		LOG << "GoalStraightPosPlan::after some lines" <<endl;
		Point point = wm->getBody().getPos().asPoint();
		try {
				Point homePoint = library.positioning.getHomePoint(wm->getBody().getUniNum());
				LOG << "Home Point is : " << homePoint << endl;
				if (wm->getBall().getPos().getX() < 44)
				{
					if (isInPlayers(wm->getBody().getUniNum() , "AB")) {
						homePoint.y *= config["GoalStraightPos"]["BasicGeometry"]["ABYFactor"].asFloat();	
						if (homePoint.y > 30)
								homePoint.y = 30;
						if (homePoint.y < -30)
								homePoint.y = -30;
						throw homePoint;
					}
				}
				
				if (wm->getBody().getDistance(homePoint) > 7.5) {
						LOG << "going to home point..." << endl;
						throw homePoint;
				}
				LOG << "oppOffsideLine.x : " << wm->getOppOffsideLine(true) << endl;
				if (wm->getOppOffsideLine(true) < wm->getBody().getPos().getX()) {
						LOG << "warning...  running away from offside..." << endl;
						throw Point(wm->getOppOffsideLine(true) - 1, wm->getBody().getPos().getY());
				}
				if (wm->getBody().getDistance(wm->getBall()) > 15 && 
					getProperAngle(wm->getBody().getBodyDir(), 0) > PATH_ABS_ANGLE)
				{
					LOG << "urgent turn to goal:1" << endl;
					turnToGoal();
				}
				if (wm->getBody().getDistance(wm->getBall()) > 5 && 
					getProperAngle(wm->getBody().getBodyDir(), 0) > 100)
				{
					LOG << "urgent urgent turn to goal: 2" << endl;
					turnToGoal();
				}
				Vector moveDir = getBestVectorInFreeArea();
				if (moveDir.getMagnitude() == 0) {
					turnToGoal();
				} else {
					moveDir += wm->getBody().getPos();
					throw moveDir.asPoint();
				}
    }
    catch (Point point)
    {
		bool posOnlyFlag = false;
		if (wm->getBallStatus() == BS_FREE_BALL_TMM || wm->getBallStatus() == BS_KICKABLE_TMM_OPP)
		{
			if (wm->getCurInterCalculate().existFastestPlayer()) {
				LOG << "Fastest OK ... " << endl;
				LOG << "P:" << wm->getCurInterCalculate().getFastestInterPlayer().getInterPoint() << endl;
				if (wm->getCurInterCalculate().getFastestInterPlayer().getInterPoint().x < 35)
					posOnlyFlag = true;
			}
		}
		if (wm->getOppOffsideLine(true) < point.x) {
			LOG << "not going to offside...." << endl;
			turnToGoal();
		}
		Command * command = NULL;
        LOG << "sending command..." << endl;
		if (point.getDistance(wm->getBody().getPos().asPoint()) > 5.0  || posOnlyFlag) {
			LOG << "-----------------> Positioning " << endl;
			command =  library.positioning.getPositioningCommand(point);
		} else {
			LOG << "--------------> PGTP" << endl;
			PreciseGotoPoint pgtp(AT_OFFENSE, point, 0.5, wm->getBody(), true, 20);
			command = pgtp.getCommand();
		}
		if (!command) 
			command = new EmptyCommand(AT_OFFENSE);
        command->setCreator(AT_OFFENSE);
        throw command;
    }

    LOG << "GoalStraightPosPlan Now Kill Program" << endl;
    assert(0);
}


float
GoalStraightPosPlan::getValueForPoint(const Point & calcPoint) 
{
		float goalieDist = -1;
//		float dist = 0;
		Vector v1, v2, v;
#ifdef LOG_INDIVIDUAL_POINT
		LOG << "Calc point:: " << calcPoint.x << ' ' << calcPoint.y;
#endif
		v1.setByPoints(calcPoint, Point(52.5, +6.3));
		v2.setByPoints(calcPoint, Point(52.5, -6.3));
		v.setByPoints(calcPoint, calcPoint);
		bool goalieFound = false;
		Player ballOwner = wm->getNearestTmmToBall();

		for (int i =0 ; i <  HALF_PLAYERS_NUM; ++i)
				if (wm->getHalfPlayer(TID_OPPONENT, i).isGoalie()) {
						goalieFound = true;
						v.setByPoints (calcPoint, wm->getHalfPlayer(TID_OPPONENT,
																												i).getPos().asPoint());
				}
		for (int i =0 ; i <  FULL_PLAYERS_NUM; ++i)
				if (wm->getFullPlayer(TID_OPPONENT, i).isGoalie()) {
						goalieFound = true;
						v.setByPoints (calcPoint, wm->getFullPlayer(TID_OPPONENT,
																												i).getPos().asPoint());
				}
		LOG << " golieFound " << goalieFound;
		float alpha = 0.0;
		float d1 = normalizeAngle(v1.getDirection()),
				d2 = normalizeAngle(v2.getDirection());
		if (goalieFound) {
				float d = normalizeAngle(v.getDirection());
				LOG << "d1: " << d1 << "d2: " << d2 << "d: " << d << endl;				
				if ((d1 >= d) && (d >= d2)) {
						// positioning is OK...
						float pa1 = getProperAngle (d1,d);
						float pa2 = getProperAngle (d,d2);
						if (pa1 > pa2) 
								alpha = pa1;
						else
								alpha = pa2;
				}
				else {
						alpha = getProperAngle(d1,d2);
				}
				goalieDist = v.getMagnitude();
		}
		else {
				LOG << "Warning.. goalie not found. (not isValid checked)" << endl;
				alpha = 0.0;
				goalieDist = 0.0;
		}
				
		Point homePoint = library.positioning.getHomePoint(wm->getBody().getUniNum());
		if (isInPlayers(wm->getBody().getUniNum(), "AB"))
			homePoint.y *= config["GoalStraightPos"]["BasicGeometry"]["ABYFactor"].asFloat();
		if (homePoint.y > 30)
			homePoint.y = 30;
		if (homePoint.y < -30)
			homePoint.y = -30;
		float realDFH = calcPoint.getDistance(homePoint);
		float distFromHome = 0.0;
		if (realDFH > 5.0) {
			distFromHome = exp(realDFH - 5.0) /
				 config["GoalStraightPos"]["Factors"]["StabilityPower"].asFloat();
		}

#ifdef LOG_INDIVIDUAL_POINT
		LOG << " g:" << goalieDist << " alpha:" << alpha;
#endif

		float sigmaM = 0.0;
		for (int i =0 ; i <  FULL_PLAYERS_NUM; ++i)
				if (wm->getFullPlayer(TID_OPPONENT, i).isValid()) {
						if (calcPoint.getDistance(wm->getFullPlayer(TID_OPPONENT, i).getPos().asPoint()) < 
								config["GoalStraightPos"]["Factors"]["SemiMarkArea"].asFloat()) {
								sigmaM -= 1/(calcPoint.getDistance(wm->getFullPlayer(TID_OPPONENT, i).getPos().asPoint())+1.0);
						}
						if (wm->getMarkers().find(wm->getFullPlayer(TID_OPPONENT, i).getUniNum())
								!= wm->getMarkers().end()) {
#ifdef LOG_INDIVIDUAL_POINT
								LOG << "cmp.. " <<
										wm->getFullPlayer(TID_OPPONENT, i).getUniNum()
										<< ".. ";
#endif
								sigmaM -= config["GoalStraightPos"]["Factors"]["MarkedPlayersExtra"].asFloat() *
										1/(calcPoint.getDistance(wm->getFullPlayer(TID_OPPONENT, i).getPos().asPoint())+1.0);
						}
				}

		for (int i =0 ; i <  HALF_PLAYERS_NUM; ++i)
				if (wm->getHalfPlayer(TID_OPPONENT, i).isValid()) {
						if (calcPoint.getDistance(wm->getHalfPlayer(TID_OPPONENT, i).getPos().asPoint()) < 
								config["GoalStraightPos"]["Factors"]["SemiMarkArea"].asFloat()) {
								sigmaM -= 1/(calcPoint.getDistance(wm->getHalfPlayer(TID_OPPONENT, i).getPos().asPoint())+1.0);
						}
						if (wm->getMarkers().find(wm->getHalfPlayer(TID_OPPONENT, i).getUniNum())
								!= wm->getMarkers().end()) {
#ifdef LOG_INDIVIDUAL_POINT
								LOG << "chmplayer.... " <<
										wm->getHalfPlayer(TID_OPPONENT, i).getUniNum()
										<< endl;
#endif
								sigmaM -= config["GoalStraightPos"]["Factors"]["MarkedPlayersExtra"].asFloat() *
										1/(calcPoint.getDistance(wm->getHalfPlayer(TID_OPPONENT, i).getPos().asPoint())+1.0);
						}
				}
#ifdef LOG_INDIVIDUAL_POINT
		LOG << " sigmaM " << sigmaM << " dfHome " << distFromHome << endl;
#endif

		float value = 
				config["GoalStraightPos"]["Factors"]["Alpha"].asFloat() *alpha  + 
				config["GoalStraightPos"]["Factors"]["GoalieDist"].asFloat()*goalieDist + 
				config["GoalStraightPos"]["Factors"]["SigmaM"].asFloat()*sigmaM +
				config["GoalStraightPos"]["Factors"]["Stability"].asFloat() * distFromHome;
		return value;
					
}
				
				
float GoalStraightPosPlan::calculatePath(Vector moveVector, int numPoints,
										 float decay)
{
	Body pathPlayer(wm->getBody());
	Command *pathCommand;
	Point targetPoint;
	float weight, decayMultiplier;
	int i;
	weight = 0;
	float value;
	float oppOffsideX;
	decayMultiplier = 1;
	oppOffsideX = wm->getOppOffsideLine(true);
	
	for (i = 0; i < numPoints; i++) {			
			targetPoint= pathPlayer.getPos() + moveVector;
			
			if ( (abs(targetPoint.x) > 52.5) || (abs(targetPoint.y) > 34)) {
					LOG << "skipping out.." << endl;
					continue;
			}
			LOG << "s1.7:" << i << endl;
			if (targetPoint.x > oppOffsideX) {
					LOG << "offside... giving minus weight" << endl;
					weight += config["GoalStraightPos"]["Factors"]["OffsideWeight"].asFloat()
									* decayMultiplier;
			}
			
			pathCommand = PreciseGotoPoint(AT_NONE,targetPoint, 0.5, pathPlayer, true, 5).getCommand();
			pathPlayer.simulateByAction(pathCommand);
			pathPlayer.simulateByDynamics();
			oppOffsideX += 0.5; // <--== BARRASI
			delete pathCommand;
			value = getValueForPoint(pathPlayer.getPos()); 
			LOG << "GoalStraightPosPlan:CP mv=" << moveVector <<
					" point=" << pathPlayer.getPos() << " value=" << value << endl;
			weight += value * decayMultiplier;
			decayMultiplier *= decay;
	}
	LOG << "GoalStraightPosPlan:W " << weight << endl;
	return weight;
}

Vector
GoalStraightPosPlan::getBestVectorInFreeArea() 
{
	int pathPoints = config["GoalStraightPos"]["BasicGeometry"]["PathPoints"].asInt();
	float pathDecay = config["GoalStraightPos"]["BasicGeometry"]["PathDecay"].asFloat();

    LOG << "calculating required and sufficient parameters...." << endl;
		Point bestPoint;
		vector <Vector> directionVector;
		Vector tempVector;
		if (isAngleOK(wm->getBody().getBodyDir())) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir()); // go forward
			directionVector.push_back(tempVector);	
		}
		if (isAngleOK(wm->getBody().getBodyDir() + 180)) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir() + 180); // go backward
			directionVector.push_back(tempVector);	
		}
		if (isAngleOK(wm->getBody().getBodyDir()) + 30) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir() + 30); // go +30
			directionVector.push_back(tempVector);	
		}
		if (isAngleOK(wm->getBody().getBodyDir()) + 30 + 180) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir() + 30 + 180); // go +30
			directionVector.push_back(tempVector);	
		}
		if (isAngleOK(wm->getBody().getBodyDir()) - 30) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir() - 30); // go -30
			directionVector.push_back(tempVector);	
		}
		if (isAngleOK(wm->getBody().getBodyDir()) - 30 + 180) {
			tempVector.setAsPolar(2, wm->getBody().getBodyDir() - 30 + 180); // go -30 backward
			directionVector.push_back(tempVector);	
		}
		tempVector.setAsCartesian(0, 0); // wait here
		directionVector.push_back(tempVector);	
		
		float maxWeight = -10000.0;
		int maxer = 0;
		// checking moving paths...
		for (unsigned i =0 ; i < directionVector.size(); ++i) {
				float newWeight = calculatePath(directionVector[i], pathPoints, pathDecay);
				LOG << "GoalStraightPosPlan:direction[" << i << "].angle=" << directionVector[i].getDirection()
					 << " weight=" << 
						newWeight << endl;
				if (newWeight >= maxWeight) {
						maxWeight = newWeight;
						maxer = i;
				}
		}
		LOG << "Best Value: " << maxWeight << endl;
		LOG << "I say: " << directionVector[maxer] << " [" << maxer << "]" << " angle:" <<
			 directionVector[maxer].getDirection() << endl;
		return directionVector[maxer];
}

double
GoalStraightPosPlan::successRate()
{
    return 1.0;
}

bool
GoalStraightPosPlan ::isFinished()
{
    if (!isInPlayers(wm->getBody().getUniNum(), "234"))
        return true;
	if (wm->getBallStatus() == BS_KICKABLE_OPP ||
		wm->getBallStatus() == BS_FREE_BALL ||
		wm->getBallStatus() == BS_FREE_BALL_OPP)
		return true;

    if (!wm->isBallKickable() &&
        library.gwSelection == WOBS_OFFENSE)
        return false;
    return true;
}

bool
GoalStraightPosPlan::isInPenaltyArea(const Point & p) const
{
    return ((abs(p.y) <= 20) && (p.x >= 36.2));
}

float
GoalStraightPosPlan::getProperAngle(float angle1, float angle2) const
{
		return abs(normalizeAngle(angle1-angle2));
}

void
GoalStraightPosPlan::turnToGoal()
{
	TurnToAngle a(AT_OFFENSE, 0, wm->getBody());
	throw a.getCommand();
}

bool
GoalStraightPosPlan::isAngleOK(float angle) const 
{
	angle = getProperAngle(angle, 0);
	if (angle > PATH_ABS_ANGLE && angle < (180 - PATH_ABS_ANGLE))
		return false;
	else
		return true;
}

