/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Mostafa Rokooey
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#ifndef __OURTYPES_H
#define __OURTYPES_H

#include <AssignmentGraph.h>
#include <Task.h>

struct OurTypesRates
{
	float main;
	float postCB;
	float postWB;
	float postCM;
	float postWM;
	float postCF;
	float postWF;
};

class OurTypes: public Task
{
private:
	AssignmentGraph *graph;
	int playerTypes[11];
	int changePlayerTypeArray[11];
	bool ourTypesChangedFlag;
	int lastChangedTypePlayer;
	Command *getChangeCommand(int uniNum, bool isTrainer = false);
	int getStandardUniNum(int uniNum);							//  Players uniform number : 1 ... 11
	float getPlayerWeight(int uniNum, int playerType);			//  Players uniform number : 0 ... 10
	void setPlayerGraphs(int uniNum);							//  Players uniform number : 0 ... 10
	void loadOurTypesRates(void);
	float getPlayerSpeedMaxWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getStaminaIncMaxWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getPlayerDecayWeight(int uniNum, int playerType);		//  Players uniform number : 1 ... 11
	float getInertiaMomentWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getDashPowerRateWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getPlayerSizeWeight(int uniNum, int playerType);		//  Players uniform number : 1 ... 11
	float getKickableMarginWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getKickRandWeight(int uniNum, int playerType);		//  Players uniform number : 1 ... 11
	float getExtraStaminaWeight(int uniNum, int playerType);	//  Players uniform number : 1 ... 11
	float getEffortMaxWeight(int uniNum, int playerType);		//  Players uniform number : 1 ... 11
	float getEffortMinWeight(int uniNum, int playerType);		//  Players uniform number : 1 ... 11
	
	OurTypesRates playerSpeedMaxRate;
	OurTypesRates staminaIncMaxRate;
	OurTypesRates playerDecayRate;
	OurTypesRates inertiaMomentRate;
	OurTypesRates dashPowerRateRate;
	OurTypesRates playerSizeRate;
	OurTypesRates kickableMarginRate;
	OurTypesRates kickRandRate;
	OurTypesRates extraStaminaRate;
	OurTypesRates effortMaxRate;
	OurTypesRates effortMinRate;

public:
	OurTypes(const CoachWorldModel *worldModel,	
			const CoachWorldModelHistory &worldModelHistory);

	virtual bool execute(bool isTrainer);

	void setOurTypes();
	int getPlayerType(int uniNum);						//  Players uniform number : 1 ... 11

};

#endif // __OURTYPES_H
