/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Arash Rahimi, Sassan Haradji
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#ifndef __PLAN_H_
#define __PLAN_H_

#include <stack>
#include <vector>
#include <string>
#include <GroupWorks.h>
#include <WorldModel.h>

#define THRESHOLD .06

class Form;
class Library;

class Plan
{
public:
	Plan (std::string name, const WorldModel *wm, Form &form,
			const Library &library);
	virtual ~Plan();

	virtual bool isFinished();
	virtual double successRate();

	virtual void decide();

	std::string getName();
	void setName(std::string n);

	double getThreshold();
	void setThreshold(double);

	bool tryPlan(Plan *plan, double threshold = THRESHOLD);
	bool startPlan(Plan *plan, double threshold = THRESHOLD);
	void resetPlans();

	Plan* bestPlan(int num, bool deletePlans, ...);
	Plan* bestPlan(std::vector<Plan *> &plans, bool deletePlans = true);
	Plan* bestPlan(std::vector<Plan *> &plans, std::vector<float> addedSuccessRates,
			bool deletePlans = true);

protected:
	const Library &library;
	Form &form;

	std::stack <Plan *>  subPlans;
	double threshold;
	const WorldModel *wm;
	std::string name;

	unsigned planLogLevel;
};


#endif // __PLAN_H_
