/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#ifndef __BALL_H
#define __BALL_H


#include <Object.h>


#ifndef NULL
#	define NULL 0
#endif // NULL

#define ABVR_RADIUS 7.5


class Body;
class Command;
class WorldModel;

class Ball: public Object
{
protected:
	bool collisionFlag;
	bool absoluteUpdateFlag;
	bool updateByHearPermittedFlag;

	float seePosDeviation;
	unsigned lastABVRTime;
	Vector lastABVRPosition;

	// Local serverparams for Updates and Simulates
	float kickPowerRate;
	float kickRandFactorL;
	float kickRandFactorR;
	float stoppedBallVel;

	void executeABVR(const Body &body);

public:
	Ball();
	virtual ~Ball();

	void update(const Body &body);
	void virtualUpdate(const Body &body);
	void updateByHear(float x, float y, float velMag, float velDir,
			const Body &body);
	void simulateByAction(const Body &body, const Command *bodyCycleCommand,
			bool ballCatched = false);
	void simulateByDynamics(const Body &body,
			unsigned checkCollisionLevel = 0, BallStatus ballStatus = BS_NONE,
			unsigned curTime = 0xFFFF);
	void setServerParamVars(const Param &serverParam);
	void parseFullState(const SExpression &exp, unsigned curTime);

	// Getting functions
	bool isCollision() const;
	bool isAbsoluteUpdate() const;
	bool isUpdateByHearPermitted() const;
	float getSeePosDeviation() const;

	float getKickPowerRate() const;
	float getKickRandFactorL() const;
	float getKickRandFactorR() const;
	float getStoppedBallVel() const;

	// Setting functions
	void setCollision(bool collisionFlagArg);
	void setAbsoluteUpdate(bool absoluteUpdateFlagArg);
	void setUpdateByHearPermitted(bool updateByHearPermittedFlagArg);
	void setSeePosDeviation(float seePosDeviationArg);

};

#endif // __BALL_H
